import typer
from pathlib import Path
from typing import Optional

from aerie_cli.commands.command_context import CommandContext

app = typer.Typer()

@app.command()
def new(
        path: Path = typer.Argument(default=...),
        description: Optional[str] = typer.Option(
            None, '--description', '-d', help="Description metadata"
        ),
        public: bool = typer.Option(False, '--public', '-pub', help="Indicates a public goal visible to all users (default false)"),
        name: Optional[str] = typer.Option(
            None, '--name', '-n', help="Name of the new goal (default is the file name without extension)"
        ),
        model_id: Optional[int] = typer.Option(
            None, '--model', '-m', help="Mission model ID to associate with the scheduling goal"
        ),
        plan_id: Optional[int] = typer.Option(
            None, '--plan', '-p', help="Plan ID of the specification to add this to"
        )
):
    """Upload new scheduling goal"""

    client = CommandContext.get_client()
    filename = path.stem
    extension = path.suffix
    if name is None:
        name = filename
    upload_obj = {}
    if extension == '.ts':
        with open(path, "r") as f:
            upload_obj["definition"] = f.read()
        upload_obj["type"] = "EDSL"
    elif extension == '.jar':
        jar_id = client.upload_file(path)
        upload_obj["uploaded_jar_id"] = jar_id
        upload_obj["parameter_schema"] = {}
        upload_obj["type"] = "JAR"
    else:
        raise RuntimeError(f"Unsupported goal file extension: {extension}")
    metadata = {"name": name}
    if description is not None:
        metadata["description"] = description
    metadata["public"] = public
    if model_id is not None:
        metadata["models_using"] = {"data": {"model_id": model_id}}
    if plan_id is not None:
        spec_id = client.get_scheduling_specification_for_plan(plan_id)
        metadata["plans_using"] = {"data": {"specification_id": spec_id}}
    upload_obj["metadata"] = {"data": metadata}
    resp = client.upload_scheduling_goals([upload_obj])
    id = resp[0]["goal_id"]
    typer.echo(f"Uploaded scheduling goal to venue. ID: {id}")


@app.command()
def update(
        path: Path = typer.Argument(default=...),
        goal_id: Optional[int] = typer.Option(None, '--goal', '-g', help="Goal ID of goal to be updated (will search by name if omitted)"),
        name: Optional[str] = typer.Option(None, '--name', '-n', help="Name of the goal to be updated (ignored if goal is provided, default is the file name without extension)"),
):
    """Upload an update to a scheduling goal"""
    client = CommandContext.get_client()
    filename = path.stem
    extension = path.suffix
    if goal_id is None:
        if name is None:
            name = filename
        goal_id = client.get_goal_id_for_name(name)
    upload_obj = {"goal_id": goal_id}
    if extension == '.ts':
        with open(path, "r") as f:
            upload_obj["definition"] = f.read()
        upload_obj["type"] = "EDSL"
    elif extension == '.jar':
        jar_id = client.upload_file(path)
        upload_obj["uploaded_jar_id"] = jar_id
        upload_obj["parameter_schema"] = {}
        upload_obj["type"] = "JAR"
    else:
        raise RuntimeError(f"Unsupported goal file extension: {extension}")

    resp = client.upload_scheduling_goals([upload_obj])
    id = resp[0]["goal_id"]
    typer.echo(f"Uploaded new version of scheduling goal to venue. ID: {id}")


@app.command()
def delete(
    goal_id: int = typer.Option(
        ..., '--goal', '-g', help="Goal ID of goal to be deleted", prompt=True
    )
):
    """Delete scheduling goal"""
    client = CommandContext.get_client()    

    resp = client.delete_scheduling_goal(goal_id)
    typer.echo("Successfully deleted Goal ID: " + str(resp))

@app.command()
def delete_all_goals_for_plan(
    plan_id: int = typer.Option(
        ..., help="Plan ID", prompt=True
    ),
):
    client = CommandContext.get_client()

    specification = client.get_scheduling_specification_for_plan(plan_id)
    clear_goals = client.get_scheduling_goals_by_specification(specification)  # response is in asc order

    if len(clear_goals) == 0:  # no goals to clear
        typer.echo("No goals to delete.")
        return
    
    typer.echo("Deleting goals for Plan ID {plan}: ".format(plan=plan_id), nl=False)
    goal_ids = []
    for goal in clear_goals:
        goal_ids.append(goal["goal_metadata"]["id"])
        typer.echo(str(goal["goal_metadata"]["id"]) + " ", nl=False)
    typer.echo()

    client.delete_scheduling_goals(goal_ids)



import typer
from aerie_cli.aerie_client import AerieClient
from aerie_cli.utils.sessions import get_active_session_client, start_session_from_configuration
from aerie_cli.aerie_host import AerieHostConfiguration

app = typer.Typer()

class CommandContext:
    hasura_admin_secret: str = None
    alternate_configuration: AerieHostConfiguration = None

    def __init__(self) -> None:
        raise NotImplementedError

    @classmethod
    def get_client(cls) -> AerieClient:
        """Get the AerieClient for any command's execution.
        If an alternate configuration has been specified, this method will attempt to find the persistent configuration or load a file with that name. If no alternate configuration is specified, then the active session is used.
        Returns:
            AerieClient
        """
        # If the configuration was set in the CLI by the user,
        # then the returned client will be derived from that configuration.
        client = None
        if cls.alternate_configuration != None and client == None:
            session = start_session_from_configuration(cls.alternate_configuration)
            client = AerieClient(session)

        if client == None:
            # no configuration specified in CLI, so the active session will be used instead
            client = get_active_session_client()

        if cls.hasura_admin_secret:
            if client.aerie_host.aerie_jwt is None:
                raise RuntimeError(f"Unauthenticated Aerie session")
            client.aerie_host.session.headers["x-hasura-admin-secret"] = cls.hasura_admin_secret
            client.aerie_host.session.headers["x-hasura-role"] = "aerie_admin"
            client.aerie_host.session.headers["x-hasura-user-id"] = client.aerie_host.aerie_jwt.username

        return client



"""
API dataclasses emulate the structure of data for exchange with the Aerie GraphQL API.

"create" dataclasses model data for upload and "read" dataclasses model data for download.
"""

from datetime import timedelta
from typing import Any
from typing import Dict
from typing import Optional
from typing import List
from typing import Union

from attrs import define, field
from attrs import converters
from attrs import asdict
from attrs import Attribute

import arrow
from arrow import Arrow

from aerie_cli.utils.serialization import postgres_interval_to_timedelta
from aerie_cli.utils.serialization import timedelta_to_postgres_interval

import json

def convert_to_time_delta(t: Union[str, timedelta]) -> timedelta:
    """Convert a string to timedelta.

    If the input is of type timedelta, does nothing and returns the input.
    This is used in certain fields to accept both strings and timedeltas."""
    if isinstance(t, timedelta):
        return t
    return postgres_interval_to_timedelta(t)

def serialize_api(inst: type, field: Attribute, value: Any):
    if isinstance(value, timedelta):
        return timedelta_to_postgres_interval(value)
    if isinstance(value, Arrow):
        return str(value)
    return value

class ApiSerialize:
    @classmethod
    def from_dict(cls, dictionary: Dict) -> "ApiSerialize":
        return cls(**dictionary)
    def to_dict(self) -> Dict:
        return asdict(self, value_serializer=serialize_api)
    @classmethod
    def from_json(cls, dictionary: Dict) -> "ApiSerialize":
        return cls(**json.loads(dictionary))

@define
class ApiEffectiveActivityArguments(ApiSerialize):
    arguments: Dict[str, Any]
    success: bool

@define
class ActivityBase(ApiSerialize):
    """Base dataclass for an activity directive

    Fields match GraphQL field names in Aerie.
    """

    type: str
    start_offset: timedelta = field(
        converter = convert_to_time_delta
    )
    metadata: Optional[Dict[str, str]] = field(factory=lambda: {}, kw_only=True)
    name: Optional[str] = field(factory=lambda: "", kw_only=True)
    arguments: Optional[Dict[str, Any]] = field(
        factory=lambda: [], kw_only=True
    )
    anchor_id: Optional[int] = field(default=None, kw_only=True)
    anchored_to_start: Optional[bool] = field(default=None, kw_only=True)

    def __attrs_post_init__(self):

        # Enforce that anchored_to_start must be specified if an anchor ID is given
        if self.anchored_to_start is None:
            if self.anchor_id is not None:
                raise ValueError(f"anchor_id was specified but anchored_to_start was not")
            self.anchored_to_start = True


@define
class ApiActivityCreate(ActivityBase):
    """Format for uploading activity directives

    Plan ID is added because it's required to be included as part of the activity object.
    """

    plan_id: int


@define
class ApiActivityUpdate(ActivityBase):
    """Format for updating activity directives

    Plan ID is excluded as this column has restricted permissions for update
    """

    pass


@define
class ApiActivityRead(ActivityBase):
    """Format for downloading activity directives

    Downloaded activity directives have an ID which should be preserved, but is excluded in other cases (upload).
    """

    id: int

@define
class ApiActivityPlanBase(ApiSerialize):
    model_id: int
    name: str
    start_time: Arrow = field(
        converter=arrow.get
    )

@define
class ApiActivityPlanCreate(ApiActivityPlanBase):
    duration: timedelta = field(
        converter = convert_to_time_delta,
    )

@define
class ApiActivityPlanRead(ApiActivityPlanBase):
    id: int
    simulations: List[int]
    duration: timedelta = field(
        converter = convert_to_time_delta,
    )
    activity_directives: Optional[List[ApiActivityRead]] = field(
        default = None,
        converter=converters.optional(
            lambda listOfDicts: [ApiActivityRead.from_dict(d) if isinstance(d, dict) else d for d in listOfDicts])
    )
    tags: Optional[List[Dict]] = field(
        default = [], 
        converter=converters.optional(
            lambda listOfDicts: [d for d in listOfDicts]
        )
    )


@define
class ApiAsSimulatedActivity(ApiSerialize):
    type: str
    parent_id: Optional[str]
    start_timestamp: Arrow = field(
        converter = arrow.get
    )
    children: List[str]
    duration: timedelta = field(
        converter = lambda microseconds: timedelta(microseconds=microseconds)
    )
    arguments: Dict[str, Any]


@define
class ApiSimulatedResourceSample(ApiSerialize):
    x: timedelta = field(
        converter = lambda microseconds: timedelta(microseconds=microseconds)
    )
    y: Any


@define
class ApiSimulationResults(ApiSerialize):
    start: Arrow = field(
        converter = arrow.get
    )
    activities: Dict[str, ApiAsSimulatedActivity]
    unfinishedActivities: Any
    # TODO: implement constraints
    constraints: Any
    # TODO: implement events
    events: Any


@define
class ApiResourceSampleResults(ApiSerialize):
    resourceSamples: Dict[str, List[ApiSimulatedResourceSample]]


@define
class ApiMissionModelCreate(ApiSerialize):
    name: str
    version: str
    mission: str
    jar_id: str


@define
class ApiMissionModelRead(ApiMissionModelCreate):
    id: int


@define
class ApiParcelCreate(ApiSerialize):
    name: str
    command_dictionary_id: int
    channel_dictionary_id: int
    sequence_adaptation_id: int

@define
class ApiParcelRead(ApiParcelCreate):
    id: int
    parameter_dictionaries: Dict[str, int]



import json

import arrow
import typer
from rich.console import Console
from rich.table import Table

from aerie_cli.commands.command_context import CommandContext

app = typer.Typer()

def construct_constraint_metadata(name, description, model_id, plan_id):
    """
    Utility function to create metadata for a given constraint.
    @param name The constraint name
    @param description The constraint description
    @param model_id The model ID that will be using this constraint
    @param plan_id The plan ID that will be using this constraint
    """
    metadata = {
        "data": {
            "name": name, 
            "description": description
        }
    }
    if model_id != None:
        if model_id is int:
            metadata["data"]["models_using"] = {
                "data": {
                    "model_id": model_id
                }
            }
        elif model_id is list:
            metadata["data"]["models_using"] = model_id
    if plan_id != None:
        if plan_id is int:
            metadata["data"]["plans_using"] = {
                "data": {
                    "plan_id": plan_id
                }
            }
        elif plan_id is list:
            metadata["data"]["plans_using"] = plan_id
    return metadata

@app.command()
def upload(
    model_id: int = typer.Option(None, help="The model id associated with the constraint (do not input plan id)"),
    plan_id: int = typer.Option(None, help="The plan id associated with the constraint (do not input model id)"),
    name: str = typer.Option(..., help="The name of the constraint", prompt=True),
    description: str = typer.Option("", help="The description of the constraint"), # optional
    constraint_file: str = typer.Option(..., help="The file that holds the constraint", prompt=True)
):
    """Upload a constraint"""

    if model_id is None and plan_id is None:
        model_id = typer.prompt("Model id")

    client = CommandContext.get_client()
    with open(constraint_file) as in_file:
        contents = in_file.readlines()
        str_contents = " ".join(contents) 
    constraint = {
        "definition": str_contents,
        "metadata": construct_constraint_metadata(name, description, model_id, plan_id)
    }
    constraint_id = client.upload_constraint(constraint)
    typer.echo(f"Created constraint: {constraint_id}")


@app.command()
def add_to_plan(
    plan_id: int = typer.Option(None, help="The plan id associated with the constraint (do not input model id)", prompt=True),
    constraint_id: int = typer.Option(..., help="The id of the constraint", prompt=True)
):
    """Associate a constraint with a plan's constraint specification."""

    client = CommandContext.get_client()
    client.add_constraint_to_plan(constraint_id=constraint_id, plan_id=plan_id)
    typer.echo(f"Added constraint: {constraint_id} to plan: {plan_id} constraint specification")


@app.command()
def delete(
    id: int = typer.Option(..., help="The constraint id to be deleted", prompt=True)
):
    """Delete a constraint"""

    client = CommandContext.get_client()
    client.delete_constraint(id)
    typer.echo(f"Successfully deleted constraint {id}")

@app.command()
def update(
    id: int = typer.Option(..., help="The constraint id to be modified", prompt=True),
    constraint_file: str = typer.Option(..., help="The new constraint for the id", prompt=True)
):
    """Update a constraint"""
    
    client = CommandContext.get_client()
    constraint = client.get_constraint_by_id(id)
    with open(constraint_file) as in_file:
        contents = in_file.readlines()
        str_contents = " ".join(contents)
    response = client.update_constraint(id, str_contents)
    for r in response:
        typer.echo(f"Updated constraint: {r['returning']}")
    
@app.command()
def violations(
    plan_id: int = typer.Option(..., help="The plan id for the violation", prompt=True)
):

    client = CommandContext.get_client()
    constraint_violations = client.get_constraint_violations(plan_id)
    typer.echo(f"Constraint violations: {constraint_violations}")
    





from typing import List

import typer


def select_from_list(options: List[str], prompt: str = 'Select an option'):
    while True:
        for i, c in enumerate(options):
            print(f"\t{i+1}) {c}")
        choice_id = typer.prompt(prompt)
        try:
            return options[int(choice_id)-1]
        except (KeyError, ValueError):
            pass



import sys

if sys.version_info < (3, 8):
    # compatibility for python <3.8
    import importlib_metadata
else:
    from importlib import metadata as importlib_metadata

__version__ = importlib_metadata.version(__package__)



from aerie_cli.aerie_host import AerieHostConfiguration
from aerie_cli.persistent import PersistentConfigurationManager

import json

def find_configuration(configuration_identifier: str) -> AerieHostConfiguration:
    """Find a configuration by name or path.
    
    configuration_identifier (str): The name or path of a configuration. Paths should be to a configuration json file.

    Raises:
        ValueError
        FileNotFoundError

    Returns:
        AerieHostConfiguration
    """
    # search for configuration by name
    for persistent_configuration in PersistentConfigurationManager.get_configurations():
        if persistent_configuration.name != configuration_identifier:
            continue
        return persistent_configuration

    # search for configuration by path
    try:
        with open(configuration_identifier, 'r') as fid:
            try:
                found_configuration = json.load(fid)
            except ValueError as e:
                raise ValueError("File provided could not be converted to json")
    except FileNotFoundError as e:
        raise FileNotFoundError(f"No configuration exists with the path or name {configuration_identifier}")
    
    return AerieHostConfiguration.from_dict(found_configuration)


"""
Client dataclasses store data in accessible formats and provide helper methods to convert to/from the API dataclasses.
"""

from datetime import timedelta
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
from enum import Enum

from attrs import define, field
from attrs import converters
import arrow
from arrow import Arrow
import json
from attrs import asdict

from aerie_cli.utils.serialization import parse_timedelta_str
from aerie_cli.schemas.api import ApiActivityCreate
from aerie_cli.schemas.api import ApiActivityUpdate
from aerie_cli.schemas.api import ApiActivityPlanCreate
from aerie_cli.schemas.api import ApiActivityPlanRead
from aerie_cli.schemas.api import ApiActivityRead
from aerie_cli.schemas.api import ApiAsSimulatedActivity
from aerie_cli.schemas.api import ApiResourceSampleResults
from aerie_cli.schemas.api import ApiSimulatedResourceSample
from aerie_cli.schemas.api import ApiSimulationResults
from aerie_cli.schemas.api import ActivityBase
from aerie_cli.schemas.api import ApiParcelRead
from aerie_cli.schemas.api import ApiParcelCreate

def parse_timedelta_str_converter(t) -> timedelta:
    if isinstance(t, str):
        return parse_timedelta_str(t)
    if isinstance(t, timedelta):
        return t
    raise TypeError(f"{type(t)} is not a supported input. Must be str or timedelta!")

def serialize_timedelta_to_str(inst, field, value):
    if isinstance(value, timedelta):
        return value.__str__()
    if isinstance(value, Arrow):
        return str(value)
    return value

class ClientSerialize:
    @classmethod
    def from_dict(cls, dictionary: dict) -> "ClientSerialize":
        return cls(**dictionary)
    def to_dict(self) -> dict:
        return asdict(self, value_serializer=serialize_timedelta_to_str)
    @classmethod
    def from_json(cls, dictionary: dict) -> "ClientSerialize":
        return cls(**json.loads(dictionary))
    def to_json(self, indent = 1) -> dict:
        self_as_dict = self.to_dict()
        return json.dumps(self_as_dict, indent=indent)

@define
class Activity(ActivityBase, ClientSerialize):
    """Activity Directive
    
    Dataclass designed for client-side manipulation of activity directives.
    Use helper methods to covert to and from API-compatible dataclasses.
    """
    start_offset: timedelta = field(
        converter = parse_timedelta_str_converter
    )
    id: Optional[int] = field(default=None)

    def to_api_create(self, plan_id: int):
        return ApiActivityCreate(
            type=self.type,
            plan_id=plan_id,
            start_offset=self.start_offset,
            arguments=self.arguments,
            name=self.name,
            metadata=self.metadata,
            anchor_id=self.anchor_id,
            anchored_to_start=self.anchored_to_start
        )

    def to_api_update(self):
        return ApiActivityUpdate(
            type=self.type,
            start_offset=self.start_offset,
            arguments=self.arguments,
            name=self.name,
            metadata=self.metadata,
            anchor_id=self.anchor_id,
            anchored_to_start=self.anchored_to_start
        )

    @classmethod
    def from_api_read(
        cls, api_activity_read: ApiActivityRead
    ) -> "Activity":
        return Activity(
            id=api_activity_read.id,
            name=api_activity_read.name,
            type=api_activity_read.type,
            start_offset=api_activity_read.start_offset,
            arguments=api_activity_read.arguments,
            metadata=api_activity_read.metadata,
            anchor_id=api_activity_read.anchor_id,
            anchored_to_start=api_activity_read.anchored_to_start
        )

@define
class EmptyActivityPlan(ClientSerialize):
    name: str
    start_time: Arrow = field(
        converter = arrow.get
    )
    end_time: Arrow = field(
        converter = arrow.get)
    tags: Optional[List[Dict]] 

    def duration(self) -> timedelta:
        return self.end_time - self.start_time


@define
class ActivityPlanCreate(EmptyActivityPlan):
    activities: List[Activity] = field(
        converter=converters.optional(
            lambda listOfDicts: [Activity.from_dict(d) if isinstance(d, dict) else d for d in listOfDicts])
    )

    id: Optional[int] = field(
        default=None
    )
    name: Optional[str] = field(
        default=None
    )
    start_time: Optional[Arrow] = field(
        default = None,
        converter = arrow.get
    )
    end_time: Optional[Arrow] = field(
        default = None,
        converter = arrow.get
    )
    model_id: Optional[int] = field(
        default=None
    )
    sim_id: Optional[int] = field(
        default=None
    )
    tags: Optional[List[Dict]] = field(
        default=[], 
        converter=converters.optional(
            lambda listOfDicts: [d for d in listOfDicts]
        )
    )

    @classmethod
    def from_plan_read(cls, plan_read: "ActivityPlanRead") -> "ActivityPlanCreate":
        return ActivityPlanCreate(
            name=plan_read.name,
            start_time=plan_read.start_time,
            end_time=plan_read.end_time,
            activities=plan_read.activities,
            tags=plan_read.tags
        )

    def to_api_create(self, model_id: int) -> "ApiActivityPlanCreate":
        return ApiActivityPlanCreate(
            model_id=model_id,
            name=self.name,
            start_time=self.start_time,
            duration=self.end_time - self.start_time,
        )


@define
class ActivityPlanRead(EmptyActivityPlan):
    id: int
    model_id: int
    sim_id: int
    tags: Optional[List[Dict]] = field(
        default = [], 
        converter=converters.optional(
            lambda listOfDicts: [d for d in listOfDicts]
        )
    )
    activities: Optional[List[Activity]] = field(
        default = None,
        converter=converters.optional(
            lambda listOfDicts: [Activity.from_dict(d) if isinstance(d, dict) else d for d in listOfDicts])
    )

    def get_activity_start_time(self, activity: Union[int, Activity]) -> arrow.Arrow:
        """Get the effective start time of an activity instance

        Args:
            activity (Union[int, Activity]): Either the Activity Directive ID or actual object

        Returns:
            arrow.Arrow: Effective activity start time
        """

        # If an ID was given, get the activity
        if isinstance(activity, int):
            try:
                activity = next(filter(lambda a: a.id == activity, self.activities))
            except StopIteration:
                raise ValueError(f"Cannot find anchor for activity with ID {activity}")

        # If the current activity is anchored to the plan, evaluate the start time rel. to the plan
        if activity.anchor_id is None:
            if activity.anchored_to_start:
                return self.start_time + activity.start_offset
            else:
                return self.end_time + activity.start_offset

        # If the current activity is anchored to another activity, evaluate the start time rel. to that act
        if activity.anchored_to_start:
            return (
                self.get_activity_start_time(activity.anchor_id) + activity.start_offset
            )
        else:
            raise ValueError(
                f"Cannot evaluate activity start time for Activity with ID {activity.id} because it is anchored to the end of another activity"
            )

    @classmethod
    def from_api_read(cls, api_plan_read: ApiActivityPlanRead) -> "ActivityPlanRead":
        plan_start = arrow.get(api_plan_read.start_time)
        return ActivityPlanRead(
            id=api_plan_read.id,
            name=api_plan_read.name,
            model_id=api_plan_read.model_id,
            sim_id=api_plan_read.simulations[0]["id"],
            start_time=plan_start,
            end_time=plan_start + api_plan_read.duration,
            tags=api_plan_read.tags,
            activities= None if api_plan_read.activity_directives is None else [
                Activity.from_api_read(api_activity)
                for api_activity in api_plan_read.activity_directives
            ],
        )


@define
class AsSimulatedActivity(ClientSerialize):
    type: str
    id: str
    parent_id: Optional[str]
    start_time: Arrow = field(
        converter = arrow.get
    )
    children: List[str]
    duration: timedelta = field(
        converter = parse_timedelta_str_converter
    )
    parameters: Dict[str, Any]

    @classmethod
    def from_api_as_simulated_activity(
        cls, api_as_simulated_activity: ApiAsSimulatedActivity, id: str
    ):
        return AsSimulatedActivity(
            type=api_as_simulated_activity.type,
            id=id,
            parent_id=api_as_simulated_activity.parent_id,
            start_time=api_as_simulated_activity.start_timestamp,
            children=api_as_simulated_activity.children,
            duration=api_as_simulated_activity.duration,
            parameters=api_as_simulated_activity.arguments,
        )


@define
class SimulatedResourceSample(ClientSerialize):
    t: Arrow = field(
        converter = arrow.get
    )
    v: Any


@define
class SimulatedResourceTimeline(ClientSerialize):
    name: str
    values: List[SimulatedResourceSample]

    @classmethod
    def from_api_sim_res_timeline(
        cls,
        name: str,
        api_sim_res_timeline: List[ApiSimulatedResourceSample],
        profile_start_time: arrow,
    ):
        return SimulatedResourceTimeline(
            name=name,
            values=[
                SimulatedResourceSample(
                    t=profile_start_time + sample.x, v=sample.y)
                for sample in api_sim_res_timeline
            ],
        )


@define
class SimulationResults(ClientSerialize):
    start_time: Arrow = field(
        converter = arrow.get
    )
    activities: List[AsSimulatedActivity]
    resources: List[SimulatedResourceTimeline]

    @classmethod
    def from_api_results(
        cls,
        api_sim_results: ApiSimulationResults,
        api_resource_timeline: ApiResourceSampleResults,
    ):
        plan_start = api_sim_results.start
        return SimulationResults(
            start_time=plan_start,
            activities=[
                AsSimulatedActivity.from_api_as_simulated_activity(act, id)
                for id, act in api_sim_results.activities.items()
            ],
            resources=[
                SimulatedResourceTimeline.from_api_sim_res_timeline(
                    name, api_timeline, plan_start
                )
                for name, api_timeline in api_resource_timeline.resourceSamples.items()
            ],
        )


@define
class ActivityInstanceCommand(ClientSerialize):
    activity_instance_id: int
    commands: List[Dict]
    errors: List[Dict]


@define
class ExpansionRun(ClientSerialize):
    id: int
    expansion_set_id: int
    simulation_dataset_id: int
    created_at: Arrow = field(
        converter = arrow.get
    )
    activity_instance_commands: Optional[List[ActivityInstanceCommand]] = None


@define
class ExpansionSet(ClientSerialize):
    id: int
    created_at: Arrow = field(
        converter = arrow.get
    )
    updated_at: Arrow = field(
        converter = arrow.get
    )
    updated_by: str
    parcel_id: int
    expansion_rules: List[int] = field(
        converter = lambda x: [i['id'] for i in x])
    description: str
    name: str
    mission_model_id: int
    owner: str


@define
class ExpansionRule(ClientSerialize):
    id: int
    activity_type: str
    authoring_mission_model_id: int
    parcel_id: int
    expansion_logic: Optional[str] = None
    name: Optional[str] = None


@define
class ResourceType(ClientSerialize):
    name: str
    schema: Dict


@define
class ExpansionDeployRule(ClientSerialize):
    name: str
    activity_type: str
    file_name: str


@define
class ExpansionDeploySet(ClientSerialize):
    name: str
    rules: List[str]


@define
class ExpansionDeployConfiguration(ClientSerialize):
    rules: List[ExpansionDeployRule] = field(
        converter=converters.optional(
            lambda x: [ExpansionDeployRule.from_dict(d) if isinstance(d, dict) else d for d in x])
    )
    sets: List[ExpansionDeploySet] = field(
        converter=converters.optional(
            lambda x: [ExpansionDeploySet.from_dict(d) if isinstance(d, dict) else d for d in x])
    )


@define
class Parcel(ClientSerialize):
    name: str
    command_dictionary_id: int
    channel_dictionary_id: int
    sequence_adaptation_id: int
    parameter_dictionary_ids: List[int]
    id: Optional[int] = field(default=None)

    def to_api_create(self) -> ApiParcelCreate:
        return ApiParcelCreate(
            name=self.name,
            command_dictionary_id=self.command_dictionary_id,
            channel_dictionary_id=self.channel_dictionary_id,
            sequence_adaptation_id=self.sequence_adaptation_id
        )

    @classmethod
    def from_api_read(cls, api_parcel: ApiParcelRead) -> "Parcel":
        return cls(
            name=api_parcel.name,
            command_dictionary_id=api_parcel.command_dictionary_id,
            channel_dictionary_id=api_parcel.channel_dictionary_id,
            sequence_adaptation_id=api_parcel.sequence_adaptation_id,
            parameter_dictionary_ids=[p["parameter_dictionary_id"] for p in api_parcel.parameter_dictionaries],
            id = api_parcel.id
        )

class DictionaryType(Enum):
    COMMAND = 'COMMAND'
    CHANNEL = 'CHANNEL'
    PARAMETER = 'PARAMETER'


@define
class DictionaryMetadata(ClientSerialize):
    id: int
    mission: str
    version: str
    created_at: Arrow = field(
        converter=arrow.get
    )
    updated_at: Arrow = field(
        converter=arrow.get
    )


@define
class SequenceAdaptationMetadata(ClientSerialize):
    id: int
    owner: str
    updated_by: str
    created_at: Arrow = field(
        converter=arrow.get
    )
    updated_at: Arrow = field(
        converter=arrow.get
    )



import json
import typer
from typing import List, Dict
from pathlib import Path
import fnmatch

import arrow

from rich.console import Console
from rich.table import Table

from aerie_cli.commands.command_context import CommandContext
from aerie_cli.utils.prompts import select_from_list
from aerie_cli.schemas.client import ExpansionRun, ExpansionDeployConfiguration

app = typer.Typer()
sequences_app = typer.Typer()
runs_app = typer.Typer()
sets_app = typer.Typer()
app.add_typer(sequences_app, name='sequences', help='Commands for sequences')
app.add_typer(runs_app, name='runs', help='Commands for expansion runs')
app.add_typer(sets_app, name='sets', help='Commands for expansion sets')

# === Bulk Deploy Command ===

@app.command('deploy')
def bulk_deploy(
    model_id: int = typer.Option(
        ..., '--model-id', '-m', prompt='Mission Model ID',
        help='Mission Model ID'
    ),
    parcel_id: int = typer.Option(
        ..., '--parcel-id', '-p', prompt='Parcel ID',
        help='Parcel ID'
    ),
    config_file: str = typer.Option(
        ..., "--config-file", "-c", prompt="Configuration file",
        help="Deploy configuration JSON file"
    ),
    rules_path: Path = typer.Option(
        Path.cwd(), help="Path to folder containing expansion rule files"
    ),
    time_tag: bool = typer.Option(False, help="Append time tags to create unique expansion rule/set names")
):
    """
    Bulk deploy command expansion rules and sets to an Aerie instance according to a JSON configuration file.

    The configuration file contains a list of rules and a list of sets:

    ```
    {
        "rules": [...],
        "sets": [...]
    }
    ```

    Each rule must provide a unique rule name, the activity type name, and the name of the file with expansion logic:

    ```
    {
        "name": "Expansion Rule Name",
        "activity_type": "Activity Type Name",
        "file_name": "my_file.ts"
    }
    ```

    Each set must provide a unique set name and a list of rule names to add:

    ```
    {
        "name": "Expansion Set Name",
        "rules": ["Expansion Rule Name", ...]
    }
    ```
    """

    client = CommandContext.get_client()

    with open(Path(config_file), "r") as fid:
        configuration: ExpansionDeployConfiguration = ExpansionDeployConfiguration.from_dict(json.load(fid))

    name_suffix = arrow.utcnow().format("_YYYY-MM-DDTHH-mm-ss") if time_tag else ""

    # Loop and upload all expansion rules
    uploaded_rules = {}
    for rule in configuration.rules:
        try:
            with open(rules_path.joinpath(rule.file_name), "r") as fid:
                expansion_logic = fid.read()

            rule_id = client.create_expansion_rule(
                expansion_logic=expansion_logic,
                activity_name=rule.activity_type,
                model_id=model_id,
                parcel_id=parcel_id,
                name=rule.name + name_suffix
            )
            typer.echo(f"Created expansion rule {rule.name + name_suffix}: {rule_id}")
            uploaded_rules[rule.name] = rule_id
        except:
            typer.echo(f"Failed to create expansion rule {rule.name}")

    for set in configuration.sets:
        try:
            rule_ids = []
            for rule_name in set.rules:
                if rule_name in uploaded_rules.keys():
                    rule_ids.append(uploaded_rules[rule_name])
                else:
                    typer.echo(f"No uploaded rule {rule_name} for set {set.name}")
            
            assert len(rule_ids)

            set_id = client.create_expansion_set(
                parcel_id=parcel_id,
                model_id=model_id,
                expansion_ids=rule_ids,
                name=set.name + name_suffix
            )

            typer.echo(f"Created expansion set {set.name + name_suffix}: {set_id}")
        except:
            typer.echo(f"Failed to create expansion set {set.name}")


# === Commands for expansion runs ===


@runs_app.command('create')
def expand_simulation(
    simulation_dataset_id: str = typer.Option(
        ..., '--sim-id', '-s', prompt='Simulation Dataset ID',
        help='Simulation Dataset ID'
    ),
    expansion_set_id: str = typer.Option(
        ..., '--expansion-set', '-e', prompt='Expansion Set ID',
        help='Expansion Set ID'
    )
):
    """
    Run command expansion on a simulation dataset
    """
    client = CommandContext.get_client()

    expansion_run_id = client.expand_simulation(
        int(simulation_dataset_id), int(expansion_set_id))

    print(f"Expansion Run ID: {expansion_run_id}")


@runs_app.command('list')
def list_expansion_runs(
    plan_id: str = typer.Option(
        None, '--plan-id', '-p',
        help='Plan ID'
    ),
    simulation_dataset_id: str = typer.Option(
        None, '--sim-id', '-s',
        help='Simulation Dataset ID'
    )
):
    """
    List expansion runs for a given plan or simulation dataset

    Runs are listed in reverse chronological order.
    """

    if simulation_dataset_id is None and plan_id is None:
        choice = select_from_list(
            ['Plan ID', 'Simulation Dataset ID'], 'Choose one to specify')
        if choice == 'Plan ID':
            plan_id = typer.prompt('Enter Plan ID')
        elif choice == 'Simulation Dataset ID':
            simulation_dataset_id = typer.prompt('Enter Simulation Dataset ID')

    client = CommandContext.get_client()

    if simulation_dataset_id is None:
        simulation_datasets = client.get_simulation_dataset_ids_by_plan_id(
            plan_id)
        table_caption = f'All runs for Plan ID {plan_id}'
    else:
        simulation_datasets = [simulation_dataset_id]
        table_caption = f'All runs for Simulation Dataset ID {simulation_dataset_id}'

    runs: List[ExpansionRun] = []
    for sd in simulation_datasets:
        runs += client.list_expansion_runs(sd)

    table = Table(
        title="Expansion Runs",
        caption=table_caption
    )
    table.add_column("Run ID", no_wrap=True)
    table.add_column("Expansion Set", no_wrap=True)
    table.add_column("Simulation Dataset ID", no_wrap=True)
    table.add_column("Creation Time", no_wrap=True)
    for run in runs:
        table.add_row(
            str(run.id),
            str(run.expansion_set_id),
            str(run.simulation_dataset_id),
            run.created_at.ctime()
        )

    Console().print(table)

# === Commands for sequences ===


@sequences_app.command('list')
def list_sequences(
    plan_id: str = typer.Option(
        None, '--plan-id', '-p',
        help='Plan ID'
    ),
    simulation_dataset_id: str = typer.Option(
        None, '--sim-id', '-s',
        help='Simulation Dataset ID'
    )
):
    """
    List sequences for a given plan or simulation dataset

    Sequences are listed in alphabetical order.
    """

    if simulation_dataset_id is None and plan_id is None:
        choice = select_from_list(
            ['Plan ID', 'Simulation Dataset ID'], 'Choose one to specify')
        if choice == 'Plan ID':
            plan_id = typer.prompt('Enter Plan ID')
        elif choice == 'Simulation Dataset ID':
            simulation_dataset_id = typer.prompt('Enter Simulation Dataset ID')

    client = CommandContext.get_client()

    if simulation_dataset_id is None:
        simulation_datasets = client.get_simulation_dataset_ids_by_plan_id(
            plan_id)
        table_caption = f'All sequences for Plan ID {plan_id}'
    else:
        simulation_datasets = [simulation_dataset_id]
        table_caption = f'All sequences for Simulation Dataset ID {simulation_dataset_id}'

    seq_ids_by_dataset: Dict[List[str]] = {}
    for sd in simulation_datasets:
        seq_ids_by_dataset[sd] = client.list_sequences(sd)

    table = Table(
        title="Sequences",
        caption=table_caption
    )
    table.add_column("Sequence ID", no_wrap=True)
    table.add_column("Simulation Dataset ID", no_wrap=True)
    for sim_dataset, seq_ids in seq_ids_by_dataset.items():
        for seq_id in seq_ids:
            table.add_row(
                seq_id,
                str(sim_dataset)
            )

    Console().print(table)


@sequences_app.command('create')
def create_sequence(
    simulation_dataset_id: str = typer.Option(
        ..., '--sim-id', '-s', prompt='Simulation Dataset ID',
        help='Simulation Dataset ID'
    ),
    seq_id: str = typer.Option(
        ..., '--sequence-id', '-t', prompt='Sequence ID',
        help='Sequence ID to download'
    )
):
    """
    Create a sequence on a simulation dataset
    """
    client = CommandContext.get_client()
    existing_sequences = client.list_sequences(simulation_dataset_id)
    if seq_id in existing_sequences:
        Console().print(f"Sequence already exists: {seq_id}", style='red')
        return

    client.create_sequence(seq_id, simulation_dataset_id)
    Console().print(f"Successfully created sequence: {seq_id}", style='green')

    selection_methods = [
        'Select activity types by glob string',
        'Add all activities from expansion run',
        'Blank sequence (no activities)'
    ]

    selection_method = select_from_list(
        selection_methods,
        'Choose how to add simulated activities to the sequence'
    )

    if selection_methods.index(selection_method) == 0:
        match_str = typer.prompt('Enter glob string')
        sim_results = client.get_simulation_results(simulation_dataset_id)
        ids_to_add = []
        for act in sim_results:
            if len(fnmatch.filter([act['activity_type_name']], match_str)):
                ids_to_add.append(act['id'])

    elif selection_methods.index(selection_method) == 1:
        ids_to_add = client.get_simulated_activity_ids(simulation_dataset_id)

    elif selection_methods.index(selection_method) == 2:
        ids_to_add = []

    if len(ids_to_add):
        client.link_activities_to_sequence(
            seq_id, simulation_dataset_id, ids_to_add)

        Console().print(f"Added activities to {seq_id}", style='green')
    else:
        Console().print("No activities added to sequence.", style='green')


@sequences_app.command('delete')
def delete_sequence(
    simulation_dataset_id: str = typer.Option(
        ..., '--sim-id', '-s', prompt='Simulation Dataset ID',
        help='Simulation Dataset ID'
    ),
    seq_id: str = typer.Option(
        ..., '--sequence-id', '-t', prompt='Sequence ID',
        help='Sequence ID to download'
    )
):
    """
    Delete sequence on a simulation dataset
    """
    CommandContext.get_client().delete_sequence(
        seq_id, simulation_dataset_id)

    Console().print(f"Successfully deleted sequence: {seq_id}", style='green')


@sequences_app.command('download')
def download_sequence(
    simulation_dataset_id: str = typer.Option(
        ..., '--sim-id', '-s', prompt='Simulation Dataset ID',
        help='Simulation Dataset ID'
    ),
    seq_id: str = typer.Option(
        ..., '--sequence-id', '-t', prompt='Sequence ID',
        help='Sequence ID to download'
    ),
    output_fn: str = typer.Option(
        ..., '--output', '-o', prompt='Output File',
        help='Name of output JSON file'
    )
):
    """
    Download a SeqJson file from an Aerie sequence
    """
    if not output_fn.endswith('.json'):
        output_fn == '.json'
    output_fn = Path(output_fn)

    client = CommandContext.get_client()

    seq_dict = client.get_expanded_sequence(seq_id, int(simulation_dataset_id))

    with open(output_fn, 'w') as fid:
        json.dump(seq_dict, fid, indent=2)

# === Commands for expansion sets ===


@sets_app.command('list')
def list_expansion_sets():
    """
    List all expansion sets
    """
    client = CommandContext.get_client()
    sets = client.list_expansion_sets()
    parcels = client.list_parcels()
    parcels_by_id = {p.id: p for p in parcels}

    table = Table(
        title="Expansion Sets"
    )
    table.add_column("ID", no_wrap=True)
    table.add_column("Name", no_wrap=True)
    table.add_column("Model ID", no_wrap=True)
    table.add_column("Parcel", no_wrap=True)
    table.add_column("Owner", no_wrap=True)
    for e_set in sets:
        table.add_row(
            str(e_set.id),
            str(e_set.name),
            str(e_set.mission_model_id),
            parcels_by_id[e_set.parcel_id].name,
            e_set.owner
        )

    Console().print(table)


@sets_app.command('get')
def get_expansion_set(
    expansion_set_id: str = typer.Option(
        ..., '--expansion-set', '-e', prompt='Expansion Set ID',
        help='Expansion Set ID'
    )
):
    """
    View all rules in an expansion set
    """
    client = CommandContext.get_client()
    sets = client.list_expansion_sets()
    try:
        set = next(filter(lambda s: s.id == int(expansion_set_id), sets))
    except StopIteration:
        Console().print(
            f"No expansion set with ID: {expansion_set_id}", style='red')
        return

    rules = client.list_expansion_rules()
    rules = list(filter(lambda r: r.id in set.expansion_rules, rules))

    table = Table(title=f"Expansion Set {expansion_set_id} Contents")
    table.add_column("Rule ID", no_wrap=True)
    table.add_column("Activity Type", no_wrap=True)
    for r in rules:
        table.add_row(
            str(r.id),
            r.activity_type
        )
    Console().print(table)


@sets_app.command('create')
def create_expansion_set(
    model_id: int = typer.Option(
        ..., '--model-id', '-m', prompt='Mission Model ID',
        help='Mission Model ID'
    ),
    parcel_id: int = typer.Option(
        ..., '--parcel-id', '-p', prompt='Parcel ID',
        help='Parcel ID'
    ),
    name: str = typer.Option(
        ..., "--name", "-n", prompt="Expansion set name",
        help="Expansion set name"
    ),
    activity_types: List[str] = typer.Option(
        None, '--activity-types', '-a',
        help='Activity types to be included in the set'
    ),
    rule_ids: List[int] = typer.Option(
        None, '--rule-ids', '-r',
        help='Expansion rules to be included in the set'
    )
):
    """
    Create an expansion set

    Uses the newest expansion rules for each given activity type.
    Specify either a list of activity type names or rule IDs. If activity type 
    names are given, rules are filtered to only those designated for the given 
    mission model and sequencing parcel, then the highest rule ID is used for 
    each activity type named.
    """
    client = CommandContext.get_client()
    expansion_rules = client.get_rules_by_type()

    if activity_types and rule_ids:
        Console().print(
            "Only specify either activity type name(s) or rule ID(s).", style='red')
        return

    if not activity_types and not rule_ids:
        choice = select_from_list(
            ["By name", "By expansion rule ID"], "Choose how to specify rules")
        if choice == "By name":
            types_str = typer.prompt('Activity Types (separate with spaces)')
            activity_types = types_str.split(' ')
        else:
            rule_ids_str = typer.prompt(
                'Expansion rule IDs (separate with spaces)')
            rule_ids = [int(r) for r in rule_ids_str.split(' ')]

    if activity_types:

        rule_ids = []
        for at in activity_types:

            try:
                at_rules = expansion_rules[at]
                for r in at_rules:
                    if not r.parcel_id == parcel_id:
                        at_rules.remove(r)
                        continue
                    if not r.authoring_mission_model_id == model_id:
                        at_rules.remove(r)
                        continue
                rule_ids.append(max([r.id for r in at_rules]))
            except (KeyError, ValueError) as e:
                print(e)
                Console().print(
                    f"No expansion rules for activity type: {at}", style='red')
                return

    set_id = client.create_expansion_set(
        parcel_id, model_id, rule_ids, name)
    Console().print(f"Created expansion set: {set_id}")



"""Main CLI application

`app` is the CLI application with which all commands, subcommands, and callbacks are registered.
"""
import typer
from typing import Optional

from aerie_cli.commands import models
from aerie_cli.commands import plans
from aerie_cli.commands import configurations
from aerie_cli.commands import expansion
from aerie_cli.commands import constraints
from aerie_cli.commands import scheduling
from aerie_cli.commands import metadata

from aerie_cli.commands.command_context import CommandContext
from aerie_cli.__version__ import __version__
from aerie_cli.persistent import (
    PersistentConfigurationManager,
    PersistentSessionManager,
)
from aerie_cli.utils.prompts import select_from_list
from aerie_cli.utils.sessions import (
    start_session_from_configuration,
    get_active_session_client,
)
from aerie_cli.utils.configurations import find_configuration

app = typer.Typer()
app.add_typer(plans.plans_app, name="plans")
app.add_typer(models.app, name="models")
app.add_typer(configurations.app, name="configurations")
app.add_typer(expansion.app, name="expansion")
app.add_typer(constraints.app, name="constraints")
app.add_typer(scheduling.app, name="scheduling")
app.add_typer(metadata.app, name="metadata")


def print_version(print_version: bool):
    if print_version:
        typer.echo(__version__)
        raise typer.Exit()


def set_alternate_configuration(configuration_identifier: str):
    if configuration_identifier == None:
        return
    found_configuration = find_configuration(configuration_identifier)

    CommandContext.alternate_configuration = found_configuration


def setup_global_command_context(hasura_admin_secret: str):
    CommandContext.hasura_admin_secret = hasura_admin_secret


@app.callback()
def app_callback(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        callback=print_version,
        help="Print Aerie-CLI package version and exit.",
    ),
    hasura_admin_secret=typer.Option(
        default="",
        help="Hasura admin secret that will be put in the header of graphql requests.",
    ),
    configuration=typer.Option(
        None,
        "--configuration",
        "-c",
        callback=set_alternate_configuration,
        help="Set a configuration to use rather than the persistent configuration.\n\
            Accepts either a configuration name or the path to a configuration json.\n\
            Configuration names are prioritized over paths.",
    ),
):
    setup_global_command_context(hasura_admin_secret)


@app.command("activate")
def activate_session(
    name: str = typer.Option(
        None, "--name", "-n", help="Name for this configuration", metavar="NAME"
    ),
    username: str = typer.Option(
        None, "--username", "-u", help="Specify/override configured Aerie username", metavar="USERNAME"
    ),
    role: str = typer.Option(
        None, "--role", "-r", help="Specify a non-default role", metavar="ROLE"
    ),
    force: bool = typer.Option(False, "--force", help="Force connection to Aerie host and ignore version compatibility")
):
    """
    Activate a session with an Aerie host using a given configuration
    """
    if name is None:
        name = select_from_list(
            [c.name for c in PersistentConfigurationManager.get_configurations()]
        )

    conf = PersistentConfigurationManager.get_configuration_by_name(name)

    session = start_session_from_configuration(conf, username, force=force)

    if role is not None:
        if role in session.aerie_jwt.allowed_roles:
            session.change_role(role)
        else:
            typer.echo(f"Role {role} not in allowed roles")

    PersistentSessionManager.set_active_session(session)


@app.command("deactivate")
def deactivate_session():
    """
    Deactivate any active session
    """
    name = PersistentSessionManager.unset_active_session()
    if name is None:
        typer.echo("No active session")
    else:
        typer.echo(f"Deactivated session: {name}")


@app.command("role")
def change_role(
    role: str = typer.Option(
        None, "--role", "-r", help="New role to selec", metavar="ROLE"
    )
):
    """
    Change Aerie permissions role for the active session
    """
    client = get_active_session_client()

    if role is None:
        typer.echo(f"Active Role: {client.aerie_host.active_role}")
        role = select_from_list(client.aerie_host.aerie_jwt.allowed_roles)

    client.aerie_host.change_role(role)

    PersistentSessionManager.set_active_session(client.aerie_host)

    typer.echo(f"Changed role to: {client.aerie_host.active_role}")


@app.command("status")
def print_status():
    """
    Returns information about the current Aerie session.
    """

    client = CommandContext.get_client()

    if client.aerie_host.configuration_name:
        typer.echo(f"Active configuration: {client.aerie_host.configuration_name}")

    typer.echo(f"Active role: {client.aerie_host.active_role}")



import json

import arrow
import typer
from rich.console import Console
from rich.table import Table

from aerie_cli.commands.command_context import CommandContext

app = typer.Typer()


@app.command()
def upload(
    mission_model_path: str = typer.Option(
        ...,
        "--mission-model-path",
        "-i",
        help="The input file from which to create an Aerie model",
        prompt=True,
    ),
    model_name: str = typer.Option(
        ..., "--model-name", "-n", help="Name of mission model", prompt=True
    ),
    version: str = typer.Option(
        "",
        "--version",
        "-v",
        help="Mission model verison",
        show_default=True,
        prompt=True,
    ),
    time_tag_version: bool = typer.Option(
        False, "--time-tag-version", help="Use timestamp for model version"
    ),
    sim_template: str = typer.Option(
        "", help="Simulation template file", show_default=True
    ),
):
    """Upload a single mission model from a .jar file."""
    # Determine Aerie UI model version
    if time_tag_version:
        version = arrow.utcnow().isoformat()

    # Initialize Aerie client
    client = CommandContext.get_client()

    # Upload mission model file to Aerie server
    model_id = client.upload_mission_model(
        mission_model_path=mission_model_path,
        project_name=model_name,
        mission="",
        version=version,
    )

    if sim_template != "":
        # Get file name
        name = sim_template.split(".")[0]

        # Read args as json
        with open(sim_template) as in_file:
            json_obj = json.load(in_file)

        # Attach sim template to model
        client.upload_sim_template(model_id=model_id, args=json_obj, name=name)
        typer.echo(f"Attached simulation template to model {model_id}.")

    typer.echo(f"Created new mission model: {model_name} with Model ID: {model_id}")


@app.command()
def delete(
    model_id: int = typer.Option(
        ..., "--model-id", "-m", help="Mission model ID to be deleted", prompt=True
    ),
):
    """Delete a mission model by its model id."""

    model_name = CommandContext.get_client().delete_mission_model(model_id)
    typer.echo(f"Mission Model `{model_name}` with ID: {model_id} has been removed.")


@app.command()
def clean():
    """Delete all mission models."""
    client = CommandContext.get_client()

    resp = client.get_mission_models()
    for api_mission_model in resp:
        client.delete_mission_model(api_mission_model.id)

    typer.echo(f"All mission models have been deleted")


@app.command()
def list():
    """List uploaded mission models."""

    resp = CommandContext.get_client().get_mission_models()

    # Create output table
    table = Table(title="Current Mission Models")
    table.add_column("Model ID", style="magenta")
    table.add_column("Model Name", no_wrap=True)
    table.add_column("Model Version", no_wrap=True)
    for api_mission_model in resp:
        table.add_row(
            str(api_mission_model.id),
            str(api_mission_model.name),
            str(api_mission_model.version),
        )

    console = Console()
    console.print(table)



"""persistent.py

Manage persistent storage of Aerie host configurations and active sessions
"""

from copy import deepcopy
from pathlib import Path
from typing import List
import json
import shutil
import pickle
from datetime import datetime, timedelta, timezone

from appdirs import AppDirs

from aerie_cli.aerie_host import AerieHost, AerieHostConfiguration

# TODO add app version s.t. changes to configuration formats can be managed
APP_DIRS = AppDirs('aerie_cli')
CONFIGURATION_FILE_DIRECTORY = Path(
    APP_DIRS.user_config_dir).resolve().absolute()
CONFIGURATION_FILE_PATH = CONFIGURATION_FILE_DIRECTORY.joinpath('config.json')

SESSION_FILE_DIRECTORY = Path(APP_DIRS.user_config_dir).resolve().absolute()
SESSION_TIMESTAMP_FSTRING = r'%Y-%jT%H-%M-%S.%f'
SESSION_TIMEOUT = timedelta(hours=12)


def delete_all_persistent_files():
    shutil.rmtree(CONFIGURATION_FILE_DIRECTORY, ignore_errors=True)
    shutil.rmtree(SESSION_FILE_DIRECTORY, ignore_errors=True)


class PersistentConfigurationManager:
    _configurations: List[AerieHostConfiguration] = None

    def __init__(self) -> None:
        """Pseudo-singleton"""
        raise NotImplementedError

    @classmethod
    def _initialize(cls) -> None:
        if cls._configurations is None:
            cls.read_configurations()

    @classmethod
    def get_configurations(cls) -> List[AerieHostConfiguration]:
        cls._initialize()
        return cls._configurations

    @classmethod
    def get_configuration_by_name(cls, configuration_name: str) -> AerieHostConfiguration:
        cls._initialize()
        try:
            return next(filter(lambda c: c.name == configuration_name, cls._configurations))
        except StopIteration:
            raise ValueError(f"Unknown configuration: {configuration_name}")

    @classmethod
    def create_configuration(cls, configuration: AerieHostConfiguration) -> None:
        cls._initialize()
        if configuration.name in [c.name for c in cls._configurations]:
            raise ValueError(
                f"Configuration already exists: {configuration.name}")

        cls._configurations.append(configuration)
        cls.write_configurations()

    @classmethod
    def update_configuration(cls, configuration: AerieHostConfiguration) -> None:
        cls._initialize()
        old_configuration = cls.get_configuration_by_name(configuration.name)
        idx = cls._configurations.index(old_configuration)
        cls._configurations[idx] = (configuration)
        cls.write_configurations()

    @classmethod
    def delete_configuration(cls, configuration_name: str) -> None:
        cls._initialize()
        old_configuration = cls.get_configuration_by_name(configuration_name)
        cls._configurations.remove(old_configuration)
        cls.write_configurations()

    @classmethod
    def write_configurations(cls) -> None:
        confs = [c.to_dict() for c in cls._configurations]
        CONFIGURATION_FILE_DIRECTORY.mkdir(exist_ok=True, parents=True)
        with open(CONFIGURATION_FILE_PATH, 'w') as fid:
            json.dump(confs, fid, indent=2)

    @classmethod
    def read_configurations(cls) -> None:
        CONFIGURATION_FILE_DIRECTORY.mkdir(exist_ok=True, parents=True)
        if CONFIGURATION_FILE_PATH.is_file():
            with open(CONFIGURATION_FILE_PATH, 'r') as fid:
                try:
                    raw_confs = json.load(fid)
                except json.JSONDecodeError:
                    raise RuntimeError(
                        f"Unable to read configuration file: {str(CONFIGURATION_FILE_PATH)}")
            cls._configurations = [
                AerieHostConfiguration.from_dict(c) for c in raw_confs]
        else:
            cls._configurations = []

    @classmethod
    def reset(cls) -> None:
        cls._configurations = None


class PersistentSessionManager:
    _active_session = None

    def __init__(self) -> None:
        """Pseudo-singleton"""
        raise NotImplementedError

    @classmethod
    def _load_active_session(cls) -> None:

        if cls._active_session is not None:
            return

        # Get any/all open sessions. List in chronological order, newest first
        session_files: List[Path] = [
            f for f in SESSION_FILE_DIRECTORY.glob('*.aerie_cli.session')]
        session_files.sort(reverse=True)

        if not len(session_files):
            raise NoActiveSessionError

        # Delete any sessions older than the newest
        for session_file in session_files[1:]:
            session_file.unlink()

        # Get timestamp of newest session
        session_file = session_files[0]
        try:
            t = datetime.strptime(
                session_file.name[:-(len('.aerie_cli.session'))], SESSION_TIMESTAMP_FSTRING)
        except Exception:
            raise RuntimeError(f"Cannot parse session timestamp: {session_file.name}. Try deactivating your session.")

        # If session hasn't been used since timeout, mark as inactive
        if (datetime.now(timezone.utc).replace(tzinfo=None) - t) > SESSION_TIMEOUT:
            session_file.unlink()
            raise NoActiveSessionError

        # Un-pickle the session
        try:
            with open(session_file, 'rb') as fid:
                session: AerieHost = pickle.load(fid)
        except Exception:
            session_file.unlink()
            raise NoActiveSessionError

        # If gateway ping fails, mark session as inactive
        if not cls.set_active_session(session):
            session_file.unlink()
            raise NoActiveSessionError

    @classmethod
    def get_active_session(cls) -> AerieHost:
        cls._load_active_session()
        return cls._active_session

    @classmethod
    def set_active_session(cls, session: AerieHost) -> bool:

        if not session.check_auth():
            return False

        session_files: List[Path] = [
            f for f in SESSION_FILE_DIRECTORY.glob('*.aerie_cli.session')]
        for session_file in session_files:
            session_file.unlink()

        cls._active_session = session

        session_file = datetime.now(timezone.utc).strftime(SESSION_TIMESTAMP_FSTRING) + '.aerie_cli.session'
        session_file = SESSION_FILE_DIRECTORY.joinpath(session_file)

        with open(session_file, 'wb') as fid:
            pickle.dump(session, fid)

        return True

    @classmethod
    def unset_active_session(cls) -> str:
        """Unset any active session

        Returns:
            str: Name of unset session if any, otherwise None
        """

        try:
            cls._load_active_session()
        except NoActiveSessionError:
            return None

        fs: List[Path] = [
            f for f in SESSION_FILE_DIRECTORY.glob('*.aerie_cli.session')]
        for fn in fs:
            fn.unlink()

        name = deepcopy(cls._active_session.configuration_name)
        cls._active_session = None
        return name

    @classmethod
    def reset(cls) -> None:
        cls._active_session = None

        # Get any/all open sessions. List in chronological order, newest first
        fs: List[Path] = [
            f for f in SESSION_FILE_DIRECTORY.glob('*.aerie_cli.session')]
        if not len(fs):
            return
        # Delete all session files
        for fn in fs:
            fn.unlink()


class NoActiveSessionError(Exception):
    pass



import sys
import time
from pathlib import Path
from typing import Dict
from typing import List
from typing import Union
from copy import deepcopy

import arrow

from .schemas.api import ApiActivityPlanRead
from .schemas.api import ApiEffectiveActivityArguments
from .schemas.api import ApiMissionModelCreate
from .schemas.api import ApiMissionModelRead
from .schemas.api import ApiResourceSampleResults
from .schemas.api import ApiParcelRead
from .schemas.client import Activity
from .schemas.client import ActivityPlanCreate
from .schemas.client import ActivityPlanRead
from .schemas.client import DictionaryMetadata
from .schemas.client import DictionaryType
from .schemas.client import SequenceAdaptationMetadata
from .schemas.client import Parcel
from .schemas.client import ExpansionRun
from .schemas.client import ExpansionRule
from .schemas.client import ExpansionSet
from .schemas.client import ResourceType
from .utils.serialization import postgres_interval_to_microseconds
from .aerie_host import AerieHost


class AerieClient:
    """Client-side behavior for aerie-cli

    Class encapsulates logic to query and send files to a given Aerie host.
    """

    def __init__(self, aerie_host: AerieHost):
        """Instantiate a client with an authenticated host session

        Args:
            aerie_host (AerieHost): Aerie host information, including authentication if necessary
        """
        self.aerie_host = aerie_host

    def get_activity_plan_by_id(self, plan_id: int, full_args: str = None) -> ActivityPlanRead:
        """Download activity plan from Aerie

        Args:
            plan_id (int): ID of the plan in Aerie
            full_args (str): comma separated list of activity types for which to
            get full arguments, otherwise only modified arguments are returned.
            Set to "true" to get full arguments for all activity types.
            Disabled if missing, None, "false", or "".

        Returns:
            ActivityPlanRead: the activity plan
        """
        query = """
        query get_plans ($plan_id: Int!) {
            plan_by_pk(id: $plan_id) {
                id
                model_id
                name
                start_time
                duration
                simulations{
                    id
                }
                tags {
                    tag {
                        id
                        name
                    }
                }
                activity_directives(order_by: { start_offset: asc }) {
                    id
                    name
                    type
                    start_offset
                    arguments
                    metadata
                    anchor_id
                    anchored_to_start
                }
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(query, plan_id=plan_id)
        api_plan = ApiActivityPlanRead.from_dict(resp)
        plan = ActivityPlanRead.from_api_read(api_plan)
        return self.__expand_activity_arguments(plan, full_args)

    def list_all_activity_plans(self) -> List[ActivityPlanRead]:
        list_all_plans_query = """
        query list_all_plans {
            plan(order_by: { id: asc }) {
                id
                model_id
                name
                start_time
                duration
                simulations{
                    id
                }
                tags {
                    tag {
                        id
                        name
                    }
                }
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(list_all_plans_query)
        activity_plans = []
        for plan in resp:
            plan = ApiActivityPlanRead.from_dict(plan)
            plan = ActivityPlanRead.from_api_read(plan)
            activity_plans.append(plan)
        return activity_plans

    def get_all_activity_plans(self, full_args: str = None) -> List[ActivityPlanRead]:
        """Get all activity plans

        Args:
            full_args (str): comma separated list of activity types for which to
            get full arguments, otherwise only modified arguments are returned.
            Set to "true" to get full arguments for all activity types.
            Disabled if missing, None, "false", or "".

        Returns:
            List[ActivityPlanRead]
        """

        # List all plans then loop to get activities from each
        plans_metadata = self.list_all_activity_plans()
        plans = [self.get_activity_plan_by_id(p.id, full_args) for p in plans_metadata]

        return plans

    def get_plan_id_by_sim_id(self, simulation_dataset_id: int) -> int:
        """Get Plan ID by Simulation Dataset ID

        Args:
            simulation_dataset_id (int): Aerie Simulation Dataset ID

        Returns:
            int: Plan ID
        """
        get_plan_id_query = """
        query PlanIdBySimDatasetId($simulation_dataset_id: Int!) {
            simulation_dataset_by_pk(id: $simulation_dataset_id) {
                simulation {
                    plan {
                        id
                    }
                }
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            get_plan_id_query,
            simulation_dataset_id=simulation_dataset_id
        )
        return resp['simulation']['plan']['id']
    
    def get_tag_id_by_name(self, tag_name: str):
        get_tags_by_name_query = """
        query GetTagByName($name: String) {
            tags(where: {name: {_eq: $name}}) {
                id
            }
        }
        """

        #make default color of tag white
        create_new_tag = """
        mutation CreateNewTag($name: String, $color: String = "#FFFFFF") {
            insert_tags_one(object: {name: $name, color: $color}) {
                id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_tags_by_name_query, 
            name=tag_name
        )

        #if a tag with the specified name exists then returns the ID, else creates a new tag with this name
        if len(resp) > 0: 
            return resp[0]["id"]
        else: 
            new_tag_resp = self.aerie_host.post_to_graphql(
                create_new_tag, 
                name=tag_name
            )

            return new_tag_resp["id"]

    def add_plan_tag(self, plan_id: int, tag_name: str):
        add_tag_to_plan = """
        mutation AddTagToPlan($plan_id: Int, $tag_id: Int) {
            insert_plan_tags(objects: {plan_id: $plan_id, tag_id: $tag_id}) {
                returning {
                    tag_id
                }
            }
        }
        """
        
        #add tag to plan
        resp = self.aerie_host.post_to_graphql(
            add_tag_to_plan, 
            plan_id=plan_id, 
            tag_id=self.get_tag_id_by_name(tag_name)
        )

        return resp['returning'][0]

    def create_activity_plan(
        self, model_id: int, plan_to_create: ActivityPlanCreate
    ) -> int:

        api_plan_create = plan_to_create.to_api_create(model_id)
        create_plan_mutation = """
        mutation CreatePlan($plan: plan_insert_input!) {
            createPlan: insert_plan_one(object: $plan) {
                id
                revision
            }
        }
        """
        plan_resp = self.aerie_host.post_to_graphql(
            create_plan_mutation,
            plan=api_plan_create.to_dict(),
        )
        plan_id = plan_resp["id"]
        plan_revision = plan_resp["revision"]

        #add plan tags if exists from plan_to_create
        for tag in plan_to_create.tags:
            self.add_plan_tag(plan_id, tag["tag"]["name"])
                
        # This loop exists to make sure all anchor IDs are updated as necessary

        # Deep copy activities so we can augment and pop from the list
        activities_to_upload = deepcopy(plan_to_create.activities)

        # Map of old to new directive IDs
        directive_id_mapping = {}

        # Boolean catches errors to avoid an infinite loop
        running = True
        while len(activities_to_upload):
            running = False

            for act in activities_to_upload:
                # If activity is anchored and the anchor isn't known yet, pass
                # If activity is anchored and the anchor is known, add it
                if act.anchor_id and act.anchor_id not in directive_id_mapping.keys():
                    continue
                else:
                    if act.anchor_id:
                        act.anchor_id = directive_id_mapping[act.anchor_id]
                    directive_id_mapping[act.id] = self.create_activity(act, plan_id)
                    activities_to_upload.remove(act)
                    running = True

            if not running:
                raise RuntimeError(
                    f"Failed to anchor activities: {', '.join([act.name for act in activities_to_upload])}"
                )

        simulation_start_time = plan_to_create.start_time.isoformat()
        simulation_end_time = plan_to_create.end_time.isoformat()
        update_simulation_mutation = """
        mutation updateSimulationBounds($plan_id: Int!, $simulation_start_time: timestamptz!, $simulation_end_time: timestamptz!) {
            update_simulation(
                where: {plan_id: {_eq: $plan_id}},
                _set: {
                    simulation_start_time: $simulation_start_time,
                    simulation_end_time: $simulation_end_time
                }
            ){
                affected_rows
            }
        }
        """
        _ = self.aerie_host.post_to_graphql(
            update_simulation_mutation,
            plan_id=plan_id,
            simulation_start_time=simulation_start_time,
            simulation_end_time=simulation_end_time
        )

        return plan_id

    def create_activity(self, activity_to_create: Activity, plan_id: int) -> int:
        api_activity_create = activity_to_create.to_api_create(plan_id)
        insert_activity_mutation = """
        mutation CreateActivity($activity: activity_directive_insert_input!) {
            createActivity: insert_activity_directive_one(object: $activity) {
                id
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            insert_activity_mutation,
            activity=api_activity_create.to_dict()
        )
        activity_id = resp["id"]

        return activity_id

    def update_activity(
        self,
        activity_id: int,
        activity_to_update: Activity,
        plan_id: int
    ) -> int:
        activity_dict: Dict = activity_to_update.to_api_update().to_dict()
        update_activity_mutation = """
        mutation UpdateActvityDirective($id: Int!, $plan_id: Int!, $activity: activity_directive_set_input!) {
            updateActivity: update_activity_directive_by_pk(
            pk_columns: { id: $id, plan_id: $plan_id }, _set: $activity
            ) {
                id
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            update_activity_mutation,
            id=activity_id,
            plan_id=plan_id,
            activity=activity_dict,
        )
        return resp["id"]

    def get_all_activity_presets(self, m_id:int) -> List:
        get_all_presets_query = """
        query ($model_id: Int!) {
            activity_presets (where: {model_id:{_eq:$model_id}}){
                id
                model_id
                name
                associated_activity_type
                arguments
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_all_presets_query,
            model_id=m_id
        )
        return resp

    def upload_activity_presets(self, upload_obj):
        upload_activity_presets_query = """
        mutation upload_presets($object: [activity_presets_insert_input!]!) {
            insert_activity_presets(objects: $object, 
                                    on_conflict: {constraint: activity_presets_model_id_associated_activity_type_name_key, 
                                                  update_columns: arguments}
                                    ) {
                returning {
                    model_id
                    id
                    associated_activity_type
                    arguments
                    name
                }
            }
        }"""

        resp = self.aerie_host.post_to_graphql(
            upload_activity_presets_query, 
            object = upload_obj
        )

        return resp["returning"]

    def simulate_plan(self, plan_id: int, poll_period: int = 5) -> int:

        simulate_query = """
        query Simulate($plan_id: Int!) {
            simulate(planId: $plan_id) {
                status
                reason
                simulationDatasetId
            }
        }
        """

        def exec_sim_query():
            return self.aerie_host.post_to_graphql(simulate_query, plan_id=plan_id)

        resp = exec_sim_query()

        nonterminal_status = ["incomplete", "pending"]
        while resp["status"] in nonterminal_status:
            time.sleep(poll_period)
            resp = exec_sim_query()

        if resp["status"] == "failed":
            sys.exit(f"Simulation failed. Response:\n{resp}")

        sim_dataset_id = resp["simulationDatasetId"]
        return sim_dataset_id

    def get_resource_timelines(self, plan_id: int):
        samples = self.get_resource_samples(self.get_simulation_dataset_ids_by_plan_id(plan_id)[0])
        api_resource_timeline = ApiResourceSampleResults.from_dict(samples)
        return api_resource_timeline

    def get_resource_samples(self, simulation_dataset_id: int, state_names: List=None):
        """Pull resource samples from a simulation dataset, optionally filtering for specific states

        Each resource's values are returned in a list of points {x: <time>, y: <value>}.
        
        Times are provided in microseconds from plan start.

        Numeric resources can be either discrete-valued or vary linearly between samples. Samples are processed such 
        that a linear interpolation between samples will always return a correct value. Two points at the same 
        timestamp indicate a discontinuity.

        Args:
            simulation_dataset_id (int)
            state_names (List, optional): List of state/resource names to pull. Defaults to None (all).

        Returns:
            Dict: Object with key "resourceSamples," the value of which is a dictionary of resource sample series keyed by resource name.
        """        

        # checks to see if user inputted specific states. If so, use this query.
        if state_names:
            resource_profile_query = """
            query GetSimulationDataset($simulation_dataset_id: Int!, $state_names: [String!]) {
                simulation_dataset_by_pk(id: $simulation_dataset_id) {
                    dataset {
                        profiles(where: { name: { _in: $state_names } }) {
                            name
                            profile_segments(order_by: { start_offset: asc }) {
                                dynamics
                                start_offset
                            }
                            type
                        }
                    }
                }
            }
            """

            resp = self.aerie_host.post_to_graphql(resource_profile_query, simulation_dataset_id=simulation_dataset_id, state_names=state_names)

        else:
            resource_profile_query = """
            query GetSimulationDataset($simulation_dataset_id: Int!) {
                simulation_dataset_by_pk(id: $simulation_dataset_id) {
                    dataset {
                        profiles {
                            name
                            profile_segments(order_by: { start_offset: asc }) {
                                dynamics
                                start_offset
                            }
                            type
                        }
                    }
                }
            }
            """
            resp = self.aerie_host.post_to_graphql(resource_profile_query, simulation_dataset_id=simulation_dataset_id)
        
        
        profiles = resp["dataset"]["profiles"]

        plan_duration_query = """
        query GetPlanDuration($plan_id: Int!) {
          plan_by_pk(id: $plan_id) {
            duration
          }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            plan_duration_query,
            plan_id=self.get_plan_id_by_sim_id(simulation_dataset_id),
        )
        duration = postgres_interval_to_microseconds(resp["duration"])

        # Parse profile segments into resource timelines
        resources = {}
        for profile in sorted(profiles, key=lambda _: _["name"]):
            name = profile["name"]
            profile_segments = profile["profile_segments"]
            profile_type = profile["type"]["type"]
            values = []

            for i in range(len(profile_segments)):
                segment = profile_segments[i]

                # The segment offset is the offset from plan start to the beginning of this segment
                segment_start_time = postgres_interval_to_microseconds(
                    segment["start_offset"]
                )

                # If this is *not* the last segment, then this segment ends where the next segment starts
                if i + 1 < len(profile_segments):
                    segment_end_time = postgres_interval_to_microseconds(
                        profile_segments[i + 1]["start_offset"]
                    )

                # If this is the last segment, then this segment ends at the end of the plan
                else:
                    segment_end_time = duration

                dynamics = segment["dynamics"]

                # Discrete profiles don't have rates
                if profile_type == 'discrete':

                    # Define points at the start and end of this profile segment
                    start_value = {
                        "x": segment_start_time,
                        "y": dynamics,
                    }
                    end_value = {
                        "x": segment_end_time,
                        "y": dynamics,
                    }

                    # Check if the previous point is identical to this one
                    if len(values) and (values[-1] == start_value):

                        # If the resource value hasn't changed, remove the previous point and extend out to the end of this profile segment
                        values.pop()
                        values.append(end_value)

                    else:

                        # If the value has changed, add points at the boundaries of this segment
                        values.append(start_value)
                        values.append(end_value)

                # Real profiles can have rates over time
                elif profile_type == 'real':

                    start_value = {
                        "x": segment_start_time,
                        "y": dynamics["initial"],
                    }

                    # If the last value is not identical to this segment's start, then add the start
                    if (len(values) and values[-1] != start_value) or (
                        len(values) == 0
                    ):
                        values.append(start_value)

                    # Add a value at the end of this segment
                    values.append(
                        {
                            "x": segment_end_time,
                            "y": dynamics["initial"]
                            + dynamics["rate"]
                            * ((segment_end_time - segment_start_time) / 1e6),
                        }
                    )

                else:
                    raise ValueError(f"Unknown resource profile type: {profile_type}")

            resources[name] = values
        return {
            "resourceSamples": resources
        }

    def get_simulation_results(self, sim_dataset_id: int) -> str:

        sim_result_query = """
        query Simulation($sim_dataset_id: Int!) {
            simulated_activity(where: { simulation_dataset_id: { _eq: $sim_dataset_id } }, order_by: { start_offset: asc }) {
                activity_type_name
                attributes
                directive_id
                duration
                end_time
                id
                start_offset
                start_time
                simulation_dataset_id
                parent_id
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            sim_result_query, sim_dataset_id=sim_dataset_id)
        return resp

    def delete_plan(self, plan_id: int) -> str:

        delete_plan_mutation = """
        mutation deletePlan($plan_id: Int!) {
            delete_plan_by_pk(id: $plan_id){
                name
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            delete_plan_mutation, plan_id=plan_id)

        return resp["name"]

    def upload_file(self, path: str) -> int:
        upload_timestamp = arrow.utcnow().isoformat()
        path_obj = Path(path)
        server_side_path = (
                path_obj.stem + "--" + upload_timestamp + path_obj.suffix
        )
        with open(path, "rb") as f:
            resp = self.aerie_host.post_to_gateway_files(
                server_side_path, f)
            return resp["id"]

    def upload_mission_model(
        self, mission_model_path: str, project_name: str, mission: str, version: str
    ) -> int:

        # Create unique jar identifier for server side
        jar_id = self.upload_file(mission_model_path)

        create_model_mutation = """
        mutation CreateModel($model: mission_model_insert_input!) {
            createModel: insert_mission_model_one(object: $model) {
                id
            }
        }
        """

        api_mission_model = ApiMissionModelCreate(
            name=project_name, mission=mission, version=version, jar_id=jar_id
        )

        resp = self.aerie_host.post_to_graphql(
            create_model_mutation, model=api_mission_model.to_dict()
        )

        return resp["id"]

    def upload_sim_template(self, model_id: int, args: str, name: str):
        sim_template_mutation = """
            mutation uploadSimTemplate($model_id: Int!,
                                    $args: jsonb!,
                                    $name: String!) {
                insert_simulation_template(objects:
                    {
                    model_id: $model_id,
                    description: $name,
                    arguments: $args
                    }) {
                    returning {
                    id
                    }
                }
            }"""

        resp = self.aerie_host.post_to_graphql(
            sim_template_mutation, model_id=model_id, args=args, name=name
        )

        # Return simulation id
        return resp["returning"][0]["id"]

    def delete_mission_model(self, model_id: int) -> str:

        delete_model_mutation = """
        mutation deleteMissionModel($model_id: Int!) {
            delete_mission_model_by_pk(id: $model_id){
                name
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            delete_model_mutation, model_id=model_id)

        return resp["name"]

    def get_mission_models(self) -> List[ApiMissionModelRead]:

        get_mission_model_query = """
        query getMissionModels {
            mission_model {
                name
                id
                version
                mission
                jar_id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(get_mission_model_query)
        api_mission_models = [
            ApiMissionModelRead.from_dict(model) for model in resp]

        return api_mission_models

    def create_config_args(self, plan_id: int, args: dict):

        update_config_arg_query = """
        mutation UpdateConfigArgument($sim_id: Int!, $args: jsonb!) {
            updateSimulation: update_simulation_by_pk(
            pk_columns: { id: $sim_id },
            _set: {
                arguments: $args
                simulation_template_id: null
            }) {
                arguments
                id
                template: simulation_template {
                    arguments
                    description
                    id
                }
            }
        }
        """

        plan = self.get_activity_plan_by_id(plan_id)
        sim_id = plan.sim_id

        resp = self.aerie_host.post_to_graphql(
            update_config_arg_query, sim_id=sim_id, args=args)

        return resp["arguments"]

    def update_config_args(self, plan_id: int, args: dict):

        update_config_arg_query = """
        mutation UpdateConfigArgument($sim_id: Int!, $args: jsonb!) {
            updateSimulation: update_simulation_by_pk(
            pk_columns: { id: $sim_id },
            _set: {
                arguments: $args
                simulation_template_id: null
            }) {
                arguments
                id
                template: simulation_template {
                    arguments
                    description
                    id
                }
            }
        }
        """
        plan = self.get_activity_plan_by_id(plan_id)
        sim_id = plan.sim_id

        # Get currently set arguments
        curr_args_dict = self.get_config(plan_id=plan_id)

        # Out of the current arguments, update provided value and keep
        # previous arguments the same
        final_args = {}
        for arg in curr_args_dict.keys():
            if args.get(arg) is None:
                final_args[arg] = curr_args_dict[arg]
            else:
                final_args[arg] = args[arg]
                args.pop(arg)

        # Add any new arguments not previously configured
        for arg in args:
            final_args[arg] = args[arg]

        resp = self.aerie_host.post_to_graphql(
            update_config_arg_query, sim_id=sim_id, args=final_args)

        return resp["arguments"]

    def get_config(self, plan_id: int):

        plan = self.get_activity_plan_by_id(plan_id)
        sim_id = plan.sim_id

        get_config_query = """
        query getConfig($sim_id: Int!) {
            simulation_by_pk(id: $sim_id) {
                arguments
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_config_query, sim_id=sim_id)

        return resp["arguments"]

    def get_activity_interface(self, activity_name: str, model_id: int) -> str:
        """Download Typescript interface for an activity from Aerie model

        Args:
            activity_name (str): Model name of the activity
            model_id (int): ID of the model in Aerie

        Returns:
            str: Contents of the interface file
        """

        get_activity_interface_query = """
        query GetActivityTypescript(
            $activity_type_name: String!
            $mission_model_id: Int!
        ) {
            getActivityTypeScript(
                activityTypeName: $activity_type_name
                missionModelId: $mission_model_id
            ) {
                typescriptFiles {
                    content
                }
                reason
                status
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            get_activity_interface_query,
            activity_type_name=activity_name,
            mission_model_id=model_id,
        )
        return data["typescriptFiles"][0]["content"]

    def create_expansion_rule(
        self,
        expansion_logic: str,
        activity_name: str,
        model_id: str,
        parcel_id: str,
        name: str = None,
        description: str = None
    ) -> int:
        """Submit expansion logic to an Aerie instance

        Args:
            expansion_logic (str): String contents of the expansion file
            activity_name (str): Name of the activity
            model_id (str): Aerie model ID
            parcel_id (str): Aerie sequencing parcel ID
            name (str, Optional): Name of the expansion rule
            description (str, Optional): Description of the expansion rule

        Returns:
            int: Expansion Rule ID in Aerie
        """

        create_expansion_logic_query = """
        mutation CreateExpansionRule($rule: expansion_rule_insert_input!) {
            createExpansionRule: insert_expansion_rule_one(object: $rule) {
                id
            }
        }
        """
        rule = {
            "activity_type": activity_name,
            "parcel_id": parcel_id,
            "authoring_mission_model_id": model_id,
            "expansion_logic": expansion_logic,
            "name": name if (name is not None) else activity_name + arrow.utcnow().format("_YYYY-MM-DDTHH-mm-ss"),
            "description": description if (description is not None) else ""
        }
        data = self.aerie_host.post_to_graphql(
            create_expansion_logic_query,
            rule=rule
        )

        return data["id"]

    def create_expansion_set(
        self, parcel_id: int, model_id: int, expansion_ids: List[int], name: str, description: str=None
    ) -> int:
        """Create an Aerie expansion set given a list of activity IDs

        Args:
            parcel_id (int): Aerie sequencing parcel ID
            model_id (int): ID of Aerie mission model
            expansion_ids (List[int]): List of expansion IDs to include in the set
            name (str): Name of the expansion set
            description (str, Optional): Freeform description field

        Returns:
            int: Expansion set ID
        """

        create_expansion_set_query = """
        mutation CreateExpansionSet(
            $expansion_ids: [Int!]!,
            $mission_model_id: Int!,
            $parcel_id: Int!,
            $name: String!,
            $description: String!
        ) {
            createExpansionSet(
                expansionIds: $expansion_ids, 
                missionModelId: $mission_model_id, 
                parcelId: $parcel_id, 
                name: $name, 
                description: $description
            ) {
                id
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            create_expansion_set_query,
            parcel_id=parcel_id,
            mission_model_id=model_id,
            expansion_ids=expansion_ids,
            name=name,
            description="" if description is None else description
        )
        return data["id"]

    def list_expansion_sets(self) -> List[ExpansionSet]:
        list_sets_query = """
        query ListExpansionSets {
            expansion_set {
                id
                created_at
                updated_at
                owner
                updated_by
                parcel_id
                expansion_rules {
                    id
                }
                description
                name
                mission_model_id
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(list_sets_query)
        return [ExpansionSet.from_dict(i) for i in resp]

    def create_sequence(self, seq_id: str, simulation_dataset_id: int) -> None:
        """Create a sequence on a given simulation dataset

        Args:
            seq_id (str): ID of the new sequence
            simulation_dataset_id (int): ID of the parent simulation dataset
        """

        create_sequence_query = """
        mutation CreateSequence(
            $seq_id: String!
            $simulation_dataset_id: Int!
        ) {
            insert_sequence_one(
                object: {
                    simulation_dataset_id: $simulation_dataset_id,
                    seq_id: $seq_id
                }
            ) {
                seq_id
            }
        }
        """
        self.aerie_host.post_to_graphql(
            create_sequence_query,
            simulation_dataset_id=simulation_dataset_id,
            seq_id=seq_id,
        )

    def get_rule_ids_by_activity_type(self, activity_type: str) -> List[int]:
        """Get ID of all expansion rules for a given activity type

        Args:
            activity_type (str): Activity type name

        Returns:
            List[int]: Expansion rule IDs, in ascending order
        """

        get_expansion_ids_query = """
        query GetExpansionLogic(
            $activity_type: String!
        ) {
            expansion_rule(
                where: {
                    activity_type: {
                        _eq: $activity_type
                    }
                }
            ){
                id
            }
            }
        """
        data = self.aerie_host.post_to_graphql(
            get_expansion_ids_query, activity_type=activity_type)
        rule_ids = [int(v["id"]) for v in data]
        rule_ids.sort()
        return rule_ids

    def list_expansion_rules(self) -> List[ExpansionRule]:
        """List all expansion rules

        Returns:
            List[ExpansionRule]
        """
        list_rules_query = """
        query ListExpansionRules {
            expansion_rule {
                activity_type
                id
                authoring_mission_model_id
                parcel_id
                name
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(list_rules_query)
        return [ExpansionRule.from_dict(r) for r in resp]

    def get_rules_by_type(self) -> Dict[str, List[ExpansionRule]]:
        """Get all expansion rules, sorted by activity type

        Returns:
            Dict[str, List[ExpansionRule]]
        """

        rules = self.list_expansion_rules()

        rules_by_type = {}
        for r in rules:
            if r.activity_type in rules_by_type.keys():
                rules_by_type[r.activity_type].append(r)
            else:
                rules_by_type[r.activity_type] = [r]

        return rules_by_type

    def get_simulation_dataset_ids_by_plan_id(self, plan_id: int) -> List[int]:
        """Get the IDs of the simulation datasets generated from a given plan

        Args:
            plan_id (int): ID of parent plan

        Returns:
            List[int]: IDs of simulation datasets in descending order
        """

        get_simulation_dataset_query = """
        query GetSimulationDatasetId($plan_id: Int!) {
          simulation(where: {plan_id: {_eq: $plan_id}}, order_by: { id: desc }, limit: 1) {
            simulation_datasets(order_by: { id: desc }) {
              id
            }
          }
        }
        """
        data = self.aerie_host.post_to_graphql(
            get_simulation_dataset_query, plan_id=plan_id)
        return [d["id"] for d in data[0]["simulation_datasets"]]

    def update_simulation_boundary(self, plan_id:int, sim_start: str, sim_end: str) -> int:
        change_boundary_mutation = """
        mutation ChangeBounds($plan_id: Int!, $new_start: timestamptz!, $new_end: timestamptz!) {
        update_simulation(where: { plan_id: { _eq: $plan_id}}, _set: { simulation_end_time: $new_end, simulation_start_time: $new_start}
        ) {
                affected_rows
                returning{    
      					simulation_end_time
      					simulation_start_time
                }
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            change_boundary_mutation,
            plan_id=plan_id,
            new_start=sim_start,
            new_end=sim_end
        )
        return str(data["affected_rows"])

    def get_simulation_boundary(self, plan_id:int) -> int:
        get_boundary_mutation = """
        query GetBounds($plan_id: Int!) {
        simulation_by_pk(id: $plan_id) {
            simulation_end_time
            simulation_start_time
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            get_boundary_mutation,
            plan_id=plan_id
        )
        return str(f"start: {data['simulation_start_time']} end: {data['simulation_end_time']}")

    def expand_simulation(
        self, simulation_dataset_id: int, expansion_set_id: int
    ) -> int:
        """Expand simulated activities from a simulation dataset given an expansion set

        Args:
            simulation_dataset_id (int): Dataset of activities to be expanded
            expansion_set_id (int): ID of expansion set to use

        Returns:
            int: Expansion Run ID
        """

        expand_simulation_query = """
        mutation ExpandPlan(
            $expansion_set_id: Int!
            $simulation_dataset_id: Int!
        ) {
            expandAllActivities(
                expansionSetId: $expansion_set_id,
                simulationDatasetId: $simulation_dataset_id
            ) {
                id
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            expand_simulation_query,
            expansion_set_id=expansion_set_id,
            simulation_dataset_id=simulation_dataset_id,
        )

        return int(data["id"])

    def list_expansion_runs(self, simulation_dataset_id: int) -> List[ExpansionRun]:
        """
        List all expansion runs from a given simulation dataset.
        """
        get_runs_query = """
        query GetExpansionRuns($simulation_dataset_id: Int!) {
            expansion_run(order_by: { created_at: desc }, where: { simulation_dataset_id: { _eq: $simulation_dataset_id } }) {
                created_at
                id
                expansion_set_id
                simulation_dataset_id
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            get_runs_query,
            simulation_dataset_id=simulation_dataset_id
        )
        return [ExpansionRun.from_dict(r) for r in resp]

    def get_expansion_run(
        self, expansion_run_id: int, include_commands: bool = False
    ) -> ExpansionRun:
        """
        Get metadata about an expansion run and, optionally, all expanded 
        activity instance commands/errors.
        """
        get_run_query = """
        query GetExpansionRun($expansion_run_id: Int!, $include_commands: Boolean!) {
            expansion_run(order_by: { created_at: desc }, where: { id: { _eq: $expansion_run_id } }) {
                created_at
                id
                expansion_set_id
                simulation_dataset_id
                activity_instance_commands @include(if: $include_commands) {
                    activity_instance_id
                    commands
                    errors
                }
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            get_run_query,
            expansion_run_id=expansion_run_id,
            include_commands=include_commands
        )
        return ExpansionRun.from_dict(resp[0])

    def link_activities_to_sequence(
        self, seq_id: str, simulation_dataset_id: int, simulated_activity_ids: List[int]
    ) -> None:
        """Link a set of simulated activities to a sequence for expansion

        Take a set of simulated activities (not plan activities) of a given
        simulation dataset and link them to an existing sequence so that
        expansion outputs from the given activities are included in the
        sequence.

        Args:
            seq_id (str): ID of the sequence to which activities will be linked
            simulation_dataset_id (int): Dataset which contains the activities being linked
            simulated_activity_ids (List[int]): IDs of simulated activities to be linked
        """

        for simulated_activity_id in simulated_activity_ids:

            link_activity_to_sequence_query = """
            mutation LinkSimulatedActivityToSequence(
                $seq_id: String!
                $simulation_dataset_id: Int!
                $simulated_activity_id: Int!
            ) {
                insert_sequence_to_simulated_activity_one(
                    object: {
                        seq_id: $seq_id
                        simulated_activity_id: $simulated_activity_id
                        simulation_dataset_id: $simulation_dataset_id
                    }
                ) {
                    seq_id
                }
            }
            """
            self.aerie_host.post_to_graphql(
                link_activity_to_sequence_query,
                seq_id=seq_id,
                simulated_activity_id=simulated_activity_id,
                simulation_dataset_id=simulation_dataset_id,
            )

    def get_simulated_activity_ids(self, simulation_dataset_id: int) -> List[int]:
        """Get the IDs of all simulated activities in a simulation dataset

        Args:
            simulation_dataset_id (int): ID of Aerie simulation dataset

        Returns:
            List[int]: List of simulated activity IDs
        """

        get_simulated_activity_ids_query = """
        query GetSimulatedActivities(
            $simulation_dataset_id: Int!
        ) {
            simulated_activity(
                where: {
                    simulation_dataset_id: { _eq: $simulation_dataset_id }
                }
            ) {
                id
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            get_simulated_activity_ids_query,
            simulation_dataset_id=simulation_dataset_id,
        )
        simulated_activity_ids = [int(o["id"]) for o in data]
        return simulated_activity_ids

    def get_expanded_sequence(self, seq_id: str, simulation_dataset_id: int) -> Dict:
        """Get SeqJson from an expanded Aerie sequence

        Args:
            seq_id (str): ID of the sequence
            simulation_dataset_id (int): ID of the simulation dataset being expanded

        Returns:
            Dict: SeqJson as Python Dictionary
        """

        get_expanded_sequence_query = """
        query GetSequenceSeqJson(
            $seq_id: String!
            $simulation_dataset_id: Int!
        ) {
            getSequenceSeqJson(
                seqId: $seq_id,
                simulationDatasetId: $simulation_dataset_id
            ) {
                seqJson
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            get_expanded_sequence_query,
            seq_id=seq_id,
            simulation_dataset_id=simulation_dataset_id,
        )
        return data["seqJson"]

    def list_sequences(self, simulation_dataset_id: int) -> List[str]:
        """List all sequences tied to a simulation dataset

        Args:
            simulation_dataset_id (int): ID on the Aerie host

        Returns:
            List[str]: Sequence IDs
        """

        list_sequences_query = """
        query ListSequences($simulation_dataset_id: Int!) {
            sequence(where: { simulation_dataset_id: { _eq: $simulation_dataset_id } }) {
                seq_id
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            list_sequences_query,
            simulation_dataset_id=simulation_dataset_id
        )
        return [s["seq_id"] for s in data]

    def delete_sequence(self, seq_id: str, simulation_dataset_id: int) -> None:
        """Delete a command sequence

        Args:
            seq_id (str): Sequence ID
            simulation_dataset_id (int): ID of Simulation Dataset where the sequence exists
        """
        delete_sequence_query = """
        mutation DeleteExpansionSequence($seq_id: String!, $simulation_dataset_id: Int!) {
            deleteExpansionSequence: delete_sequence_by_pk(seq_id: $seq_id, simulation_dataset_id: $simulation_dataset_id) {
                seq_id
            }
        }
        """
        self.aerie_host.post_to_graphql(
            delete_sequence_query,
            seq_id=seq_id,
            simulation_dataset_id=simulation_dataset_id
        )

    def get_all_expansion_run_commands(self, expansion_run_id: int) -> List:
        """Get commands from all activity instances in an expansion run

        Provides commands without linking to a particular sequence.

        Args:
            expansion_run_id (int): ID of Aerie expansion run

        Returns:
            List: SeqJson-formatted command steps
        """

        expansion_run_commands_query = """
        query GetAllExpandedCommandsForExpansionRun(
            $expansion_run_id: Int!
        ) {
            activity_instance_commands(
                where: {
                    expansion_run: {
                        id: { _eq: $expansion_run_id }
                    }
                }
            ) {
                activity_instance_id
                commands
                errors
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            expansion_run_commands_query, expansion_run_id=expansion_run_id
        )

        expansion_run_commands = []
        for activity_instance in data:
            if activity_instance["commands"]:
                expansion_run_commands.extend(activity_instance["commands"])

        return expansion_run_commands

    def get_all_activity_types(self, model_id: int) -> List[str]:
        """Get a list of all activity types defined in a given mission model

        Args:
            model_id (int): ID of the Aerie mission model

        Returns:
            List[str]: List of activity type names
        """

        get_types_query = """
        query GetActivityTypes(
            $model_id: Int!
        ) {
            activity_type(where: { model_id: { _eq: $model_id } }) {
                name
            }
        }
        """
        data = self.aerie_host.post_to_graphql(
            get_types_query, model_id=model_id)
        activity_types = [o["name"] for o in data]
        return activity_types

    def get_typescript_dictionary(self, command_dictionary_id: int) -> str:
        """Download Typescript command dictionary for writing EDSL Sequences and Expansion

        Prepends mission model name and command dictionary version in a comment
        block header of the Typescript.

        Args:
            command_dictionary_id (int): ID of command dictionary in Aerie instance

        Returns:
            str: Typescript file contents
        """

        get_command_dictionary_metadata_query = """
        query GetCommandDictionaryMetadata($command_dictionary_id: Int!) {
        command_dictionary(where: { id: { _eq: $command_dictionary_id } }) {
            mission
            version
        }
        }
        """

        get_typescript_dictionary_query = """
        query GetCommandTypescript($command_dictionary_id: Int!) {
        getCommandTypeScript(commandDictionaryId: $command_dictionary_id) {
            reason
            status
            typescriptFiles {
                content
                filePath
            }
        }
        }
        """

        data = self.aerie_host.post_to_graphql(
            get_command_dictionary_metadata_query,
            command_dictionary_id=command_dictionary_id,
        )[0]

        command_dictionary_mission = data["mission"]
        command_dictionary_version = data["version"]

        data = self.aerie_host.post_to_graphql(
            get_typescript_dictionary_query, command_dictionary_id=command_dictionary_id
        )

        typescript_dictionary_string = next(
            filter(
                lambda x: x["filePath"].endswith("command-types.ts"),
                data["typescriptFiles"],
            )
        )["content"]

        # TODO add Aerie version below once the API supports this
        # Include metadata about command dictionary in header
        typescript_dictionary_string = "\n".join(
            [
                "/**",
                f"* Mission:                    {command_dictionary_mission}",
                f"* Command Dictionary Version: R{command_dictionary_version}",
                "*/",
                "",
                typescript_dictionary_string,
            ]
        )

        return typescript_dictionary_string

    def get_scheduling_goals_by_specification(self, spec_id):
        list_all_goals_by_spec_query = """
        query ($spec: Int!){
            scheduling_specification_goals(where: {
                specification_id:{_eq:$spec}
            }){
                goal_metadata {
                    id
                    name
                    description
                    public
                    owner
                    updated_by
                    created_at
                    updated_at
                }
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(list_all_goals_by_spec_query, spec=spec_id)

        return resp

    def create_dictionary(self, dictionary: str, persist: bool=True) -> int:
        """Upload an AMPCS command, channel, or parameter dictionary to an Aerie instance

        Args:
            dictionary (str): Contents from XML dictionary file (newlne-delimited)
            persist (bool, optional): Persist the uploaded source dictionary in Aerie. Defaults to True.

        Returns:
            int: Dictionary ID
        """

        query = """
        mutation CreateDictionary($dictionary: String!, $persist: Boolean!) {
            createDictionary: uploadDictionary(dictionary: $dictionary, persistDictionaryToFilesystem: $persist) {
                command
                channel
                parameter
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            query,
            dictionary=dictionary,
            persist=persist
        )
        return next(iter(resp.values()))["id"]

    def list_dictionaries(self) -> Dict[DictionaryType, List[DictionaryMetadata]]:
        """List all command, parameter, and channel dictionaries

        Returns:
            List[DictionaryMetadata]
        """

        command_dictionaries = self.aerie_host.post_to_graphql("""query ListDictionaries {
            command_dictionary {
                id
                version
                updated_at
                created_at
                mission
            }
        }
        """)
        channel_dictionaries = self.aerie_host.post_to_graphql("""query ListDictionaries {
            channel_dictionary {
                id
                version
                updated_at
                created_at
                mission
            }
        }
        """)
        parameter_dictionaries = self.aerie_host.post_to_graphql("""query ListDictionaries {
            parameter_dictionary {
                id
                version
                updated_at
                created_at
                mission
            }
        }
        """)
        return {
            DictionaryType.COMMAND: [DictionaryMetadata.from_dict(i) for i in command_dictionaries],
            DictionaryType.CHANNEL: [DictionaryMetadata.from_dict(i) for i in channel_dictionaries],
            DictionaryType.PARAMETER: [DictionaryMetadata.from_dict(
                i) for i in parameter_dictionaries]
        }

    def delete_dictionary(self, id: int, dictionary_type: Union[str, DictionaryType]) -> None:
        """Delete AMPCS dictionary

        Args:
            id (int): _description_
            dictionary_type (Union[str, DictionaryType]): _description_
        """
        
        if not isinstance(dictionary_type, DictionaryType):
            dictionary_type = DictionaryType(dictionary_type)

        queries = {
                DictionaryType.COMMAND: """
            mutation DeleteCommandDictionary($id: Int!) {
                delete_command_dictionary_by_pk(id: $id) {
                    id
                }
            }
            """,
            DictionaryType.CHANNEL: """
            mutation DeleteChannelDictionary($id: Int!) {
                delete_channel_dictionary_by_pk(id: $id) {
                    id
                }
            }
            """,
                DictionaryType.PARAMETER: """
            mutation DeleteParameterDictionary($id: Int!) {
                delete_parameter_dictionary_by_pk(id: $id) {
                    id
                }
            }
            """
        }
        self.aerie_host.post_to_graphql(queries[dictionary_type], id=id)

    def create_sequence_adaptation(self, adaptation: str) -> int:
        """Upload Phoenix Editor sequence adaptation

        Args:
            adaptation (str): String contents of adaptation JS file

        Returns:
            int: ID of sequence adaptation
        """        
        
        query = """
        mutation CreateSequenceAdaptation($adaptation: sequence_adaptation_insert_input!) {
            createSequenceAdaptation: insert_sequence_adaptation_one(object: $adaptation) {
                id
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(query, adaptation={"adaptation": adaptation})

        return resp["id"]

    def list_sequence_adaptations(self) -> List[SequenceAdaptationMetadata]:
        """List Phoenix Editor sequence adaptations

        Returns:
            List[SequenceAdaptationMetadata]: Metadata for all existing sequence adaptations
        """
        
        query = """
        query ListSequenceAdaptations {
            sequence_adaptation {
                id
                updated_at
                created_at
                owner
                updated_by
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(query)
        return [SequenceAdaptationMetadata.from_dict(i) for i in resp]

    def update_sequence_adaptation(self, adaptation: str, id: int) -> None:
        """Update Phoenix Editor sequence adaptation

        Args:
            adaptation (str): String contents of adaptation JS file
            id (int): ID of sequence adaptation to update
        """
        
        query = """
        mutation UpdateSequenceAdaptation($adaptation: String!, $id: Int!) {
            update_sequence_adaptation_by_pk(pk_columns: {id: $id}, _set: {adaptation: $adaptation}) {
                id
            }
        }
        """
        self.aerie_host.post_to_graphql(query, adaptation={"adaptation": adaptation}, id=id)

    def delete_sequence_adaptation(self, id: int) -> None:
        """Delete Phoenix Editor sequence adaptation

        Args:
            id (int): ID of adaptation to delete
        """

        query = """
        mutation DeleteSequenceAdaptation($id: Int!) {
            delete_sequence_adaptation_by_pk(id: $id) {
                id
            }
        }
        """
        self.aerie_host.post_to_graphql(query, id=id)

    def create_parcel(self, parcel: Parcel) -> int:
        """Create sequencing parcel

        Inserts parcel entry with command dictionary, channel dictionary, and sequencing adaptation IDs.
        Links parameter dictionaries to that parcel.

        Args:
            parcel (Parcel): Contents of parcel to create

        Returns:
            int: ID of created parcel
        """

        resp = self.aerie_host.post_to_graphql(
            """
            mutation CreateParcel($parcel: parcel_insert_input!) {
                createParcel: insert_parcel_one(object: $parcel) {
                    id
                }
            }
            """, parcel=parcel.to_api_create().to_dict()
        )
        parcel_id = resp["id"]

        parameter_dictionaries = [
            {
                "parameter_dictionary_id": p,
                "parcel_id": parcel_id
            } for p in parcel.parameter_dictionary_ids
        ]
        self.aerie_host.post_to_graphql(
            """
            mutation LinkParameterDictionariesToParcel($parameter_dictionaries: [parcel_to_parameter_dictionary_insert_input!]!) {
                insert_parcel_to_parameter_dictionary(objects: $parameter_dictionaries) {
                    affected_rows
                }
            }
            """,
            parameter_dictionaries=parameter_dictionaries
        )
        return parcel_id

    def list_parcels(self) -> List[Parcel]:
        """List sequencing parcels

        Returns:
            List[Parcel]: 
        """

        resp = self.aerie_host.post_to_graphql(
            """
            query GetParcels {
                parcel {
                    channel_dictionary_id
                    command_dictionary_id
                    id
                    name
                    sequence_adaptation_id
                    parameter_dictionaries {
                        parameter_dictionary_id
                    }
                }
            }
            """
        )
        return [Parcel.from_api_read(ApiParcelRead.from_dict(p)) for p in resp]

    def delete_parcel(self, id: int) -> None:
        """Delete sequencing parcel

        Args:
            id (int): ID of parcel to delete
        """        

        self.aerie_host.post_to_graphql(
            """
            mutation DeleteParcel($id: Int!) {
                delete_parcel_by_pk(id: $id) {
                    id
                }
            }
            """,
            id=id
        )

    def upload_scheduling_goals(self, upload_object):
        """
        Bulk upload operation for uploading scheduling goals.
        @param upload_object should be JSON-like with a definition key and metadata containing the goal name and model id
        [
            {
                definition: str
                metadata: {
                    name: str,
                    models_using: {
                        data: {
                            model_id: int
                        }
                    }
                }
            },
            ...
        ]
        """
        
        upload_scheduling_goals_query = """
        mutation InsertGoal($input: [scheduling_goal_definition_insert_input!]!){
            insert_scheduling_goal_definition(objects: $input){
                returning {goal_id}
            }
        }"""
        
        resp = self.aerie_host.post_to_graphql(
            upload_scheduling_goals_query,
            input=upload_object
        )

        return resp["returning"]

    def get_scheduling_specification_for_plan(self, plan_id):
        get_scheduling_specification_for_plan_query = """
        query GetSpecificationForPlan($plan_id: Int!) {
            scheduling_specification(where: {plan_id: {_eq: $plan_id}}) {
                id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_scheduling_specification_for_plan_query, 
            plan_id=plan_id
        )
        return resp[0]["id"]

    def get_goal_id_for_name(self, name):
        get_goal_id_for_name_query = """
        query GetNameForGoalId($name: String!) {
            scheduling_goal_metadata(where: {name: {_eq: $name}}) {
                id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_goal_id_for_name_query,
            name=name
        )
        if len(resp) == 0:
            raise RuntimeError(f"No goals found with name {name}.")
        elif len(resp) > 1:
            raise RuntimeError(f"Multiple goals found with name {name}.")
        return resp[0]["id"]

    def add_goals_to_specifications(self, upload_object):
        """
        Bulk operation to add goals to specification.
        @param upload_object should be JSON-like with keys goal_id, specification_id
        [
            {
                goal_id: int,
                specification_id: int
            },
            ...
        ]
        """

        add_goal_to_specification_query = """
        mutation AddGoalToSpec($object: [scheduling_specification_goals_insert_input!]!) {
            insert_scheduling_specification_goals(objects: $object) {
                returning {
                    enabled
                    goal_id
                    priority
                    simulate_after
                    specification_id
                }
            }
        }
        """
        resp = self.aerie_host.post_to_graphql(
            add_goal_to_specification_query, 
            object = upload_object
        )

        return resp['returning']

    def delete_scheduling_goal(self, goal_id):
        return self.delete_scheduling_goals(list([goal_id]))

    def delete_scheduling_goals(self, goal_id_list):
        # We must remove the goal(s) from any specifications before deleting them
        delete_scheduling_goals_from_all_specs_query = """
        mutation DeleteSchedulingGoalsFromAllSpecs($id_list: [Int!]!) {
            delete_scheduling_model_specification_goals (where: {goal_id: {_in:$id_list}}){
                returning {goal_id}
            }
            delete_scheduling_specification_goals (where: {goal_id: {_in:$id_list}}){
                returning {goal_id}
            }
        }
        """

        resp_for_deleting_from_specs = self.aerie_host.post_to_graphql(
            delete_scheduling_goals_from_all_specs_query, 
            id_list=goal_id_list
        )

        # Note that deleting the scheduling goal metadata entry will take care of the 
        # scheduling goal definition entry too
        delete_scheduling_goals_query = """
        mutation DeleteSchedulingGoals($id_list: [Int!]!) {
            delete_scheduling_goal_metadata (where: {id: {_in:$id_list}}){
                returning {id}
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            delete_scheduling_goals_query, 
            id_list=goal_id_list
        )

        return resp["returning"]

    def get_plan_revision(self, planId):
        get_plan_revision_query = """
        query get_plan_revision($plan_id:Int) {
            plan(where: {id: {_eq: $plan_id}}){
                revision
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_plan_revision_query, 
            plan_id=planId
        )

        return resp[0]["revision"]

    def __expand_activity_arguments(self, plan: ActivityPlanRead, full_args: str = None) -> ActivityPlanRead:
        if full_args is None or full_args == "" or full_args.lower() == "false":
            return plan
        expand_all = full_args.lower() == "true"
        expand_types = {} if expand_all else set(full_args.split(","))
        for activity in plan.activities:
            if expand_all or activity.type in expand_types:
                query = """
                query ($args: ActivityArguments!, $act_type: String!, $model_id: Int!) {
                    getActivityEffectiveArguments(
                        activityArguments: $args,
                        activityTypeName: $act_type,
                        missionModelId: $model_id
                    )
                    {
                        arguments
                        success
                    }
                }
                """
                resp = self.aerie_host.post_to_graphql(
                    query,
                    args=activity.arguments,
                    act_type=activity.type,
                    model_id=plan.model_id,
                )
                activity.arguments = ApiEffectiveActivityArguments.from_dict(
                    resp).arguments
        return plan

    def upload_constraint(self, constraint):
        upload_constraint_query = """
        mutation CreateConstraint($constraint: constraint_definition_insert_input!) {
            createConstraint: insert_constraint_definition_one(object: $constraint) {
                constraint_id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(upload_constraint_query, constraint=constraint)
        return resp["constraint_id"]
    
    def add_constraint_to_plan(self, constraint_id, plan_id):
        """
        Add a constraint to a plan's constraint specification.
        @param constraint_id The constraint ID to add to the given plan
        @param plan_id The plan ID to add this constraint to
        """

        add_constraint_to_specification_query = """
            mutation InsertConstraintSpec($constraint_id: Int!, $plan_id: Int!) {
                insert_constraint_specification_one(object: {constraint_id: $constraint_id, plan_id: $plan_id}) {
                    constraint_id
                }
            }

        """
        resp = self.aerie_host.post_to_graphql(
            add_constraint_to_specification_query, 
            constraint_id = constraint_id,
            plan_id = plan_id
        )

        return resp['constraint_id']

    def delete_constraint(self, id):

        # We must remove the constraint from any specifications before deleting it
        delete_constraint_from_all_specs_query = """
        mutation DeleteConstraintsFromAllSpecs($id: Int!) {
            delete_constraint_specification(where: {constraint_id: {_eq: $id}}) {
                returning {
                    constraint_id
                    plan_id
                }
            }
        }
        """

        resp_for_deleting_from_specs = self.aerie_host.post_to_graphql(
            delete_constraint_from_all_specs_query, 
            id=id
        )

        delete_constraint_query = """
        mutation DeleteConstraint($id: Int!) {
            deleteConstraint: delete_constraint_metadata_by_pk(id: $id) {
                id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(delete_constraint_query, id=id)
        return resp["id"]
    
    def update_constraint(self, id, definition):
        old_update_constraint_query = """
        mutation UpdateConstraint($constarint_id: Int!, $constraint: constraint_definition_set_input!) {
            update_constraint_definition_by_pk(
                pk_columns: { constarint_id: $constraint_id }, _set: $constraint
            ) {
                constarint_id
                definition
                author
                created_at
            }
        }
        """

        update_constraint_query = """
        mutation UpdateConstraint($constarint_id: Int!, $definition: String!) {
            update_constraint_definition_many(
                updates: {_set: {definition: $definition}, where: {constraint_id: {_eq: $constarint_id}}}) {
                returning {
                    constraint_id
                }
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(update_constraint_query, constarint_id=id, definition=definition)
        return resp
    
    def get_constraint_by_id(self, id):
        get_constraint_by_id_query = """
        query get_constraint($constraint_id: Int!) {
            constraint_definition(where: {constraint_id: {_eq: $constraint_id}}) {
                author
                definition
                metadata {
                    description
                    name
                    plans_using {
                        plan_id
                    }
                    models_using {
                        model_id
                    }
                }
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(get_constraint_by_id_query, constraint_id=id)
        return resp
    
    def get_constraint_specification_for_plan(self, plan_id):
        get_constraint_specification_for_plan_query = """
        query GetConstraintSpecificationForPlan($plan_id: Int!) {
            constraint_specification(where: {plan_id: {_eq: $plan_id}}) {
                constraint_id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_constraint_specification_for_plan_query, 
            plan_id=plan_id
        )
        return resp[0]["id"]

    def get_constraint_violations(self, plan_id):
        get_violations_query = """
        query ($plan_id: Int!) {
            constraintResponses: constraintViolations(planId: $plan_id) {
                constraintsRun {
                    constraintId
                    constraintName
                    success
                    results {
                        resourceIds
                        gaps {
                            end
                            start
                        }
                        violations {
                            activityInstanceIds
                            windows {
                                end
                                start
                            }
                        }
                    }
                    errors {
                        message
                        stack
                        location {
                            column
                            line
                        }
                    }
                }
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(get_violations_query, plan_id=plan_id)
        return resp["constraintsRun"]

    def get_resource_types(self, model_id: int) -> List[ResourceType]:
        """Get resource types (value schema)

        Get the type information for each resource in a mission model.

        The schema format is described in the Aerie documentation [here](https://nasa-ammos.github.io/aerie-docs/mission-modeling/advanced-value-schemas/#value-schemas-in-json).

        Args:
            model_id (int): Mission model for resources

        Returns:
            List[ResourceType]
        """

        get_resource_types_query = """
        query ResourceTypes($missionModelId: Int!) {
            resourceTypes: resource_type(where: {model_id: {_eq: $missionModelId}}) {
                name
                schema
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            get_resource_types_query, missionModelId=model_id
        )
        return [ResourceType.from_dict(r) for r in resp]

    def get_directive_metadata(self) -> list:
        """Get metatdata

        Returns:
            list: a list of the metadata keys and schemas
        """
        get_metadata_query = """
        query GetMetadata {
            activity_directive_metadata_schema {
                key
                schema
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(get_metadata_query)
        return resp

    def add_directive_metadata_schemas(self, schemas: list) -> list:
        """Add metadata schemas

        The schema format should follow the documentation [here](https://nasa-ammos.github.io/aerie-docs/planning/activity-directive-metadata/).

        Args:
            schemas (list): a list of the schemas to add

        Returns:
            list: a list of the metadata keys and schemas that were added
        """
        add_schemas_query = """
        mutation CreateActivityDirectiveMetadataSchemas($schemas: [activity_directive_metadata_schema_insert_input!]!) {
            insert_activity_directive_metadata_schema(objects: $schemas) {
                returning {
                    key
                    schema
                }
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            add_schemas_query,
            schemas=schemas
        )
        return resp
        
    def delete_directive_metadata_schema(self, key) -> list:
        """Delete metadata schemas

        Returns:
            list: a list of the metadata keys that were deleted
        """
        delete_schema_query = """
        mutation DeleteDirectiveMetadataSchema($key: String!) {
            delete_activity_directive_metadata_schema_by_pk(key: $key) {
                key
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            delete_schema_query,
            key=key
        )
        return resp["key"]

    def list_plan_collaborators(self, plan_id: int) -> list:
        """List plan collaborators

        Args:
            plan_id (int): ID of Plan to list collaborators of

        Returns:
            list[str]: List of collaborator usernames
        """
        query = """
        query GetPlanCollaborators($plan_id: Int!) {
            plan_by_pk(id: $plan_id) {
                collaborators {
                    collaborator
                }
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            query,
            plan_id=plan_id
        )
        return [c["collaborator"] for c in resp["collaborators"]]

    def add_plan_collaborator(self, plan_id: int, user: str):
        """Add a plan collaborator

        Args:
            plan_id (int): ID of plan to add collaborator to
            user (str): Username of collaborator
        """
        query = """
        mutation addPlanCollaborator($plan_id: Int!, $collaborator: String!) {
            insert_plan_collaborators_one(object: {plan_id: $plan_id, collaborator: $collaborator}) {
                collaborator
            }
        }
        """

        self.aerie_host.post_to_graphql(
            query,
            plan_id=plan_id,
            collaborator=user
        )

    def delete_plan_collaborator(self, plan_id: int, user: str):
        """Delete a plan collaborator

        Args:
            plan_id (int): ID of the plan to delete a collaborator from
            user (str): Username of the collaborator
        """

        query = """
        mutation DeletePlanCollaborator($plan_id: Int!, $collaborator: String!) {
            delete_plan_collaborators_by_pk(collaborator: $collaborator, plan_id: $plan_id) {
                collaborator
                plan_id
            }
        }
        """

        resp = self.aerie_host.post_to_graphql(
            query,
            plan_id=plan_id,
            collaborator=user
        )

        if resp is None:
            raise RuntimeError(f"Failed to delete plan collaborator")



import json
import requests
from copy import deepcopy
from typing import Dict
from typing import Optional
from base64 import b64decode

from attrs import define, field

COMPATIBLE_AERIE_VERSIONS = [
    "3.6.0",
    "3.7.0",
    "3.7.1",
    "3.8.0",
    "3.8.1",
    "4.0.0",
    "4.1.0",
    "4.1.1",
    "4.2.0",
    "4.2.1"
]

class AerieHostVersionError(RuntimeError):
    pass


def process_gateway_response(resp: requests.Response) -> dict:
    """Throw a RuntimeError if the Gateway response is malformed or contains errors

    Args:
        resp (requests.Response): from a request to the gateway

    Returns:
        dict: Contents of response JSON
    """
    if not resp.ok:
        raise RuntimeError("Bad response from Aerie Gateway")

    try:
        resp_json = resp.json()
    except requests.exceptions.JSONDecodeError:
        raise RuntimeError("Bad response from Aerie Gateway")

    if "success" in resp_json.keys() and not resp_json["success"]:
        raise RuntimeError(f"Aerie Gateway request was not successful")

    return resp_json


class AerieJWT:
    def __init__(self, encoded_jwt: str) -> None:
        jwt_components = encoded_jwt.split(".")
        if not len(jwt_components) == 3:
            raise ValueError(f"Invalid JWT: {encoded_jwt}")

        encoded_jwt_payload = b64decode(jwt_components[1] + "==", validate=False)
        try:
            payload = json.loads(encoded_jwt_payload)
            self.allowed_roles = payload["https://hasura.io/jwt/claims"][
                "x-hasura-allowed-roles"
            ]
            self.default_role = payload["https://hasura.io/jwt/claims"]["x-hasura-default-role"]
            self.username = payload["username"]

        except KeyError:
            raise ValueError(f"Missing fields in JWT: {encoded_jwt}")

        except json.decoder.JSONDecodeError:
            raise ValueError(f"Cannot decode JWT: {encoded_jwt}")

        self.encoded_jwt = encoded_jwt


class AerieHost:
    """
    Abstracted interface for the Hasura and Aerie Gateway APIs of an Aerie instance.

    An instance stores the necessary URLs and authenticates using header or
    cookie information stored in the `requests.Session` object, if necessary.
    """

    def __init__(
        self,
        graphql_url: str,
        gateway_url: str,
        session: requests.Session = None,
        configuration_name: str = None,
    ) -> None:
        """

        Args:
            graphql_url (str): Route to Aerie host's GraphQL API
            gateway_url (str): Route to Aerie Gateway
            session (requests.Session, optional): Session with headers/cookies for external authentication
            configuration_name (str, optional): Name of configuration for this session
        """
        self.session = session if session else requests.Session()
        self.graphql_url = graphql_url
        self.gateway_url = gateway_url
        self.configuration_name = configuration_name
        self.aerie_jwt = None
        self.active_role = None

    def post_to_graphql(self, query: str, **kwargs) -> Dict:
        """Issue a post request to the Aerie instance GraphQL API

        Args:
            query (str): GraphQL query text
            kwargs: keyword arguments for named variables for the query

        Raises:
            RuntimeError

        Returns:
            Dict: Query response data
        """

        try:

            resp = self.session.post(
                self.graphql_url,
                json={"query": query, "variables": kwargs},
                headers=self.get_auth_headers(),
            )

            resp.raise_for_status()
            try:
                resp_json = resp.json()
            except json.decoder.JSONDecodeError:
                raise RuntimeError(f"Failed to process response")

            if "success" in resp_json.keys() and not resp_json["success"]:
                raise RuntimeError("GraphQL request was not successful")
            elif "errors" in resp_json.keys():
                raise RuntimeError(
                    f"GraphQL Error: {json.dumps(resp_json['errors'])}"
                )
            else:
                data = next(iter(resp.json()["data"].values()))

            if data is None:
                raise RuntimeError(f"Failed to process response: {resp}")

            return data

        except RuntimeError as e:
            # Re-raise with additional information
            e = str(e)

            if "password" in kwargs:

                # Remove password
                pw_sub_str = "<Password removed in aerie-cli error handling>"
                e.replace(kwargs["password"], pw_sub_str)
                kwargs = deepcopy(kwargs)
                kwargs["password"] = pw_sub_str

                # Raise exception with query, variables, and original exception
                raise RuntimeError(
                    {"query": deepcopy(query), "variables": kwargs, "exception": e}
                )

            else:

                # Decode response text if in JSON-parsable format
                try:
                    resp_contents = json.loads(resp.text)
                except json.decoder.JSONDecodeError:
                    resp_contents = resp.text

                # Raise exception with query, variables, original exception, and response
                raise RuntimeError(
                    json.dumps(
                        {
                            "query": deepcopy(query),
                            "variables": deepcopy(kwargs),
                            "exception": e,
                            "response_text": resp_contents,
                        },
                        indent=2,
                    )
                )

    def post_to_gateway_files(self, file_name: str, file_contents: bytes) -> Dict:
        """Issue a post request to upload a file via the Aerie gateway

        Args:
            file_name (str): Name of the file being uploaded
            file_contents (bytes): File contents

        Raises:
            RuntimeError: Raised if the request receives an error response

        Returns:
            Dist: JSON response
        """

        resp = self.session.post(
            self.gateway_url + "/file",
            files={"file": (file_name, file_contents)},
            headers=self.get_auth_headers(),
        )

        if resp.ok:
            return resp.json()
        else:
            raise RuntimeError(f"Error uploading file: {file_name}")

    def change_role(self, new_role: str) -> None:
        """Change role for Aerie interaction

        Args:
            new_role (str): String name of new role.
        """

        if new_role not in self.aerie_jwt.allowed_roles:
            raise ValueError(
                f"Cannot set role {new_role}. Must be one of: {', '.join(self.aerie_jwt.allowed_roles)}"
            )

        self.active_role = new_role

    def check_auth(self) -> bool:
        """Checks if session is correctly authenticated with Aerie host
        
        Looks for errors received after pinging GATEWAY_URL/auth/session.
        Also returns False if the session has not yet been authenticated (no JWT).

        Returns:
            bool: True if there were no errors in a ping against GATEWAY_URL/auth/session, False otherwise
        """
        if self.aerie_jwt is None:
            return False

        try:
            resp = self.session.get(
                self.gateway_url + "/auth/session", headers=self.get_auth_headers()
            )
        except requests.exceptions.ConnectionError:
            return False
        try:
            return resp.json()["success"]
        except Exception:
            return False

    def get_auth_headers(self):
        if self.aerie_jwt is None:
            return {}

        return {
            "Authorization": f"Bearer {self.aerie_jwt.encoded_jwt}",
            "x-hasura-role": self.active_role,
        }

    def is_auth_enabled(self) -> bool:
        """Check if authentication is enabled on an Aerie host

        Returns:
            bool: False if authentication is disabled, otherwise True
        """
        # Try to login using blank credentials. If "Authentication is disabled" is returned, we can safely skip auth
        resp = self.session.post(self.gateway_url + "/auth/login", json={"username": "", "password": ""},)

        if resp.ok:
            try:
                resp_json = process_gateway_response(resp)                
                if (
                    "message" in resp_json.keys()
                    and resp_json["message"] == "Authentication is disabled"
                ):
                    return False
            except RuntimeError:
                pass
            except requests.exceptions.JSONDecodeError:
                pass

        return True

    def authenticate(self, username: str, password: str = None, force: bool = False):

        try:
            self.check_aerie_version()
        except AerieHostVersionError as e:
            if force:
                print("Warning: " + e.args[0])
            else:
                raise

        resp = self.session.post(
            self.gateway_url + "/auth/login",
            json={"username": username, "password": password},
        )

        try:
            resp_json = process_gateway_response(resp)
        except RuntimeError:
            raise RuntimeError("Failed to authenticate")

        self.aerie_jwt = AerieJWT(resp_json["token"])
        self.active_role = self.aerie_jwt.default_role

        if not self.check_auth():
            raise RuntimeError(f"Failed to open session")

    def check_aerie_version(self) -> None:
        """Assert that the Aerie host is a compatible version

        Raises a `RuntimeError` if the host appears to be incompatible.
        """

        resp = self.session.get(self.gateway_url + "/version")

        try:
            resp_json = process_gateway_response(resp)
            host_version = resp_json["version"]
        except (RuntimeError, KeyError):
            # If the Gateway responded, the route doesn't exist
            if resp.text and "Aerie Gateway" in resp.text:
                raise AerieHostVersionError("Incompatible Aerie version: host version unknown")
            
            # Otherwise, it could just be a failed connection
            raise

        if host_version not in COMPATIBLE_AERIE_VERSIONS:
            raise AerieHostVersionError(f"Incompatible Aerie version: {host_version}")


@define
class ExternalAuthConfiguration:
    """Configure additional external authentication necessary for connecting to an Aerie host.

    Define configuration for a server which can provide additional authentication via cookies.

    auth_url (str): URL to send request for authentication
    static_post_vars (dict[str, str]): key-value constants to include in the post request variables
    secret_post_vars (list[str]): keys for secret values to include in the post request variables
    """

    auth_url: str
    static_post_vars: dict
    secret_post_vars: list

    @classmethod
    def from_dict(cls, config: Dict) -> "ExternalAuthConfiguration":
        try:
            auth_url = config["auth_url"]
            static_post_vars = config["static_post_vars"]
            secret_post_vars = config["secret_post_vars"]

        except KeyError as e:
            raise ValueError(
                f"External auth configuration missing required field: {e.args[0]}"
            )

        if not isinstance(static_post_vars, dict):
            raise ValueError(
                "Invalid value for 'static_post_vars' in external auth configuration"
            )

        if not isinstance(secret_post_vars, list):
            raise ValueError(
                "Invalid value for 'secret_post_vars' in external auth configuration"
            )

        return cls(auth_url, static_post_vars, secret_post_vars)

    def to_dict(self) -> Dict:
        return {
            "auth_url": self.auth_url,
            "static_post_vars": self.static_post_vars,
            "secret_post_vars": self.secret_post_vars,
        }


@define
class AerieHostConfiguration:
    name: str
    graphql_url: str
    gateway_url: str
    username: Optional[str] = field(default=None)
    external_auth: Optional[ExternalAuthConfiguration] = field(default=None)

    @classmethod
    def from_dict(cls, config: Dict) -> "AerieHostConfiguration":
        try:
            name = config["name"]
            graphql_url = config["graphql_url"]
            gateway_url = config["gateway_url"]
            username = config["username"] if "username" in config.keys() else None
            external_auth = ExternalAuthConfiguration.from_dict(config["external_auth"]) if "external_auth" in config.keys() else None

        except KeyError as e:
            raise ValueError(f"Configuration missing required field: {e.args[0]}")

        return cls(name, graphql_url, gateway_url, username, external_auth)

    def to_dict(self) -> Dict:
        retval = {
            "name": self.name,
            "graphql_url": self.graphql_url,
            "gateway_url": self.gateway_url
        }

        if self.username is not None:
            retval["username"] = self.username

        if self.external_auth is not None:
            retval["external_auth"] = self.external_auth.to_dict()

        return retval



import requests
from typing import Dict
import typer
from copy import deepcopy

from aerie_cli.aerie_host import AerieHost, AerieHostConfiguration, ExternalAuthConfiguration, AerieJWT
from aerie_cli.aerie_client import AerieClient
from aerie_cli.persistent import PersistentSessionManager

def get_active_session_client():
    """Instantiate AerieClient with the active host session

    Raises:
        NoActiveSessionError: if there is no active session

    Returns:
        AerieClient
    """
    session = PersistentSessionManager.get_active_session()
    return AerieClient(session)


def get_localhost_client() -> AerieClient:
    aerie_host = AerieHost("http://localhost:8080/v1/graphql", "http://localhost:9000")
    if not aerie_host.check_auth():
        raise RuntimeError(f"Failed to connect to host")
    return AerieClient(aerie_host)


def get_preauthenticated_client_native(encoded_jwt: str, graphql_url: str, gateway_url: str) -> AerieClient:
    """Get AerieClient instance preauthenticated with native Aerie JWT auth

    Args:
        encoded_jwt (str): base64-encoded Aerie JWT
        graphql_url (str) 
        gateway_url (str)

    Raises:
        RuntimeError: If connection to Aerie host fails

    Returns:
        AerieClient
    """
    aerie_host = AerieHost(graphql_url, gateway_url)
    aerie_host.aerie_jwt = AerieJWT(encoded_jwt)
    if not aerie_host.check_auth():
        raise RuntimeError(f"Failed to connect to host")
    return AerieClient(aerie_host)


def get_preauthenticated_client_cookie(cookies: dict, encoded_jwt: str, graphql_url: str, gateway_url: str) -> AerieClient:
    """Get AerieClient instance preauthenticated with an external cookie(s) and JWT

    Args:
        cookies (dict): Browser-style cookies for external authentication
        encoded_jwt (str): base64-encoded Aerie JWT
        graphql_url (str)
        gateway_url (str)

    Raises:
        RuntimeError: If connection to Aerie host fails

    Returns:
        AerieClient
    """
    session = requests.Session()
    session.cookies = requests.cookies.cookiejar_from_dict(cookies)
    aerie_host = AerieHost(graphql_url, gateway_url, session=session)
    aerie_host.aerie_jwt = AerieJWT(encoded_jwt)
    if not aerie_host.check_auth():
        raise RuntimeError(f"Failed to connect to host")
    return AerieClient(aerie_host)


def authenticate_with_external(
    configuration: ExternalAuthConfiguration, secret_post_vars: Dict[str, str] = None
) -> requests.Session:
    """Authenticate requests.Session object with an external service

    Send a post request with static and secret variables defined in `configuration` to `auth_url`.
    Cookies returned from the request are stored in the returned requests.Session object.

    Args:
        configuration (ProxyConfiguration): Proxy server configuration
        secret_post_vars (Dict[str, str], optional): Optionally provide values for some or all secret post request variable values. Defaults to None.

    Raises:
        RuntimeError: Failure to authenticate with proxy

    Returns:
        requests.Session: Session with any cookies acquired for proxy authentication
    """

    session = requests.Session()

    post_vars = deepcopy(configuration.static_post_vars)

    if secret_post_vars is None:
        secret_post_vars = {}

    for secret_var_name in configuration.secret_post_vars:
        if secret_var_name in secret_post_vars.keys():
            post_vars[secret_var_name] = secret_post_vars[secret_var_name]
        else:
            post_vars[secret_var_name] = typer.prompt(f"External authentication - {secret_var_name}", hide_input=True)

    resp = session.post(configuration.auth_url, json=post_vars)

    if not resp.ok:
        raise RuntimeError(
            f"Failed to authenticate with proxy: {configuration.auth_url}"
        )

    return session


def start_session_from_configuration(
    configuration: AerieHostConfiguration, 
    username: str = None, 
    password: str = None,
    secret_post_vars: Dict[str, str] = None,
    force: bool = False
):
    """Start and authenticate an Aerie Host session, with prompts if necessary

    If username is not provided, it will be requested via CLI prompt.

    If password is not provided but the Aerie instance has authentication enabled, it will be requested via CLI prompt.
    If the Aerie instance has authentication disabled, a password is not necessary.

    If external authentication is specified in the configuration, `secret_post_vars` can be used to pass in 
    credentials. If external auth is specified with secrets and matching credentials aren't provided, they will be 
    requested via CLI prompt.

    Args:
        configuration (AerieHostConfiguration): Configuration of host to connect
        username (str, optional): Aerie username.
        password (str, optional): Aerie password.
        secret_post_vars (Dict[str, str], optional): Optionally provide values for some or all secret post request variable values. Defaults to None.
        force (bool, optional): Force connection to Aerie host and ignore version compatibility. Defaults to False.

    Returns:
        AerieHost: 
    """

    if configuration.external_auth is None:
        session = requests.Session()
    else:
        session = authenticate_with_external(configuration.external_auth, secret_post_vars)

    hs = AerieHost(
        configuration.graphql_url,
        configuration.gateway_url,
        session,
        configuration.name,
    )

    if username is None:
        if configuration.username is None:
            username = typer.prompt("Aerie Username")
        else:
            username = configuration.username

    if password is None and hs.is_auth_enabled():
        password = typer.prompt("Aerie Password", hide_input=True)

    hs.authenticate(username, password, force)

    return hs



"""Aerie object (de)serialization utilities

Postgres interval parsing modified from Django implementation:
https://github.com/django/django/blob/0dd29209091280ccf34e07c9468746c396b7778e/django/utils/dateparse.py#L52-L64

Django is licensed under the BSD 3-Clause License:
    Copyright (c) Django Software Foundation and individual contributors.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:

        1. Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.

        2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

        3. Neither the name of Django nor the names of its contributors may be used
        to endorse or promote products derived from this software without
        specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
    ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""

import re
from datetime import timedelta


POSTGRES_INTERVAL_RE = re.compile(
    r"^"
    r"(?:(?P<days>-?\d+) (days? ?))?"
    r"(?:(?P<sign>[-+])?"
    r"(?P<hours>\d+):"
    r"(?P<minutes>\d\d):"
    r"(?P<seconds>\d\d)"
    r"(?:\.(?P<microseconds>\d{1,6}))?"
    r")?$"
)


def postgres_interval_to_timedelta(interval: str) -> timedelta:
    """Parse a Postgres inteval to a `timedelta` object

    Modified from Django interval parsing cited above.

    Args:
        interval (str): Postgres interval string

    Raises:
        ValueError: Unable to match patterns in the interval

    Returns:
        timedelta
    """
    match = POSTGRES_INTERVAL_RE.match(interval)
    if match:
        kw = match.groupdict()
        sign = -1 if kw.pop("sign", "+") == "-" else 1
        if kw.get("microseconds"):
            kw["microseconds"] = kw["microseconds"].ljust(6, "0")
        kw = {k: float(v.replace(",", ".")) for k, v in kw.items() if v is not None}
        days = timedelta(kw.pop("days", 0.0) or 0.0)
        return days + sign * timedelta(**kw)
    else:
        raise ValueError(f"Unable to parse interval string: {interval}")


def postgres_interval_to_microseconds(interval: str) -> int:
    """Convert a postgres interval string to an integer number of microseconds

    Both `timedelta` and Postgres intervals have a precision of 1us, so this conversion shouldn't introduce any rounding errors
    """
    return int(postgres_interval_to_timedelta(interval).total_seconds() * (10**6))


def timedelta_to_postgres_interval(td: timedelta) -> str:
    """Constructs a PostgreSQL interval from a `timedelta`"""
    seconds = td.days * 86400 + td.seconds
    microseconds = td.microseconds
    return f"{int(seconds)} seconds {int(microseconds)} microseconds"


def parse_timedelta_str(td_string: str) -> timedelta:
    """Parse the string format returned by timedelta.__str__"""
    pattern = r"((?P<days>[-+]?\d+)\s*day[s]?,\s*)?(?P<hours>\d{1,2}):(?P<minutes>\d{2}):(?P<seconds>\d{2})(?:\.(?P<microseconds>\d{6}))?"
    match = re.match(pattern, td_string)

    if match:
        groups = match.groupdict()
        days = int(groups["days"]) if groups["days"] else 0
        hours = int(groups["hours"])
        minutes = int(groups["minutes"])
        seconds = int(groups["seconds"])
        microseconds = int(groups["microseconds"]) if groups["microseconds"] else 0

        parsed_timedelta = timedelta(
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            microseconds=microseconds,
        )
        return parsed_timedelta

    else:
        raise ValueError(f"Invalid time string format: {td_string}")



"""configurations.py

Commands related to persistent storage of Aerie host configurations.
"""

import typer
import json
from pathlib import Path
from rich.console import Console
from rich.table import Table

from aerie_cli.aerie_host import AerieHostConfiguration
from aerie_cli.persistent import PersistentConfigurationManager, PersistentSessionManager, delete_all_persistent_files, NoActiveSessionError, CONFIGURATION_FILE_PATH
from aerie_cli.utils.prompts import select_from_list

app = typer.Typer()


@app.command('create')
def create_configuration(
    name: str = typer.Option(
        ..., prompt=True, help='Name for this configuration', metavar='NAME'),
    graphql_url: str = typer.Option(
        ..., prompt='GraphQL URL', help='URL of GraphQL API endpoint', metavar='GRAPHQL_URL'),
    gateway_url: str = typer.Option(
        ..., prompt='Gateway URL', help='URL of Aerie Gateway', metavar='GATEWAY_URL'),
    username: str = typer.Option(
        None, help='Username for authentication', metavar='USERNAME'
    )
):
    """
    Define a configuration for an Aerie host
    """
    if typer.confirm('Specify username'):
        username = typer.prompt('Username')
    else:
        username = None

    conf = AerieHostConfiguration(
        name, graphql_url, gateway_url, username)
    PersistentConfigurationManager.create_configuration(conf)


@app.command('load')
def upload_configurations(
    filename: Path = typer.Option(
        ..., '--filename', '-i', prompt='Input filename',
        help='Name of input JSON file'
    ),
    allow_overwrite: bool = typer.Option(
        False, '--allow-overwrite',
        help='Allow overwriting existing configurations'
    )
):
    """
    Load one or more configurations from a JSON file
    """
    with open(filename, 'r') as fid:
        configurations = json.load(fid)

    if isinstance(configurations, list):
        configurations = [AerieHostConfiguration.from_dict(
            c) for c in configurations]
    else:
        configurations = [AerieHostConfiguration.from_dict(configurations)]

    new_confs = []
    updated_confs = []

    for c in configurations:
        try:
            PersistentConfigurationManager.create_configuration(c)
            new_confs.append(c.name)
        except ValueError as e:
            if allow_overwrite and 'Configuration already exists' in e.args[0]:
                PersistentConfigurationManager.update_configuration(c)
                updated_confs.append(c.name)
            else:
                raise e

    if len(new_confs):
        Console().print(f"Added configurations: {', '.join(new_confs)}")

    if len(updated_confs):
        Console().print(f"Updated configurations: {', '.join(updated_confs)}")


@app.command('list')
def list_configurations():
    """
    List available Aerie host configurations
    """

    # Get the name of the active session configuration, if any
    try:
        s = PersistentSessionManager.get_active_session()
        active_config = s.configuration_name
    except NoActiveSessionError:
        active_config = None

    typer.echo(f"Configuration file location: {CONFIGURATION_FILE_PATH}")
    typer.echo()

    table = Table(title='Aerie Host Configurations',
                  caption='Active configuration in red')
    table.add_column('Host Name', no_wrap=True)
    table.add_column('GraphQL API URL', no_wrap=True)
    table.add_column('Aerie Gateway URL', no_wrap=True)
    table.add_column('Username', no_wrap=True)
    for c in PersistentConfigurationManager.get_configurations():
        if c.name == active_config:
            style = 'red'
        else:
            style = None
        table.add_row(
            c.name,
            c.graphql_url,
            c.gateway_url,
            c.username if c.username else "",
            style=style
        )

    Console().print(table)


@app.command('delete')
def delete_configuration(
    name: str = typer.Option(
        None, '--name', '-n', help='Name for this configuration', metavar='NAME', show_default=False)
):
    """
    Delete an Aerie host configuration
    """
    names = [c.name for c in PersistentConfigurationManager.get_configurations()]
    if not name:
        name = select_from_list(names)

    PersistentConfigurationManager.delete_configuration(name)

    try:
        s = PersistentSessionManager.get_active_session()
        if name == s.configuration_name:
            PersistentSessionManager.unset_active_session()
    except NoActiveSessionError:
        pass


@app.command('clean')
def delete_all_files(
    not_interactive: bool = typer.Option(
        False, help='Disable interactive prompt')
):
    """
    Remove all persistent aerie-cli files
    """
    # Please don't flame me for this double negative, it had to be done
    if not not_interactive:
        if not typer.confirm("Delete all persistent files associated with aerie-cli?"):
            return

    delete_all_persistent_files()
    # Update the PersistentConfigurationManager's cached configurations to account for the clean
    PersistentSessionManager.reset()
    PersistentConfigurationManager.reset()



"""Programmatic entrypoint for CLI application.
"""
import sys
from rich.console import Console


from aerie_cli.app import app
from aerie_cli.persistent import NoActiveSessionError
from aerie_cli.__version__ import __version__


def main():
    try:
        app()
    except NoActiveSessionError:
        Console().print(
            "There is no active session. Please start a session with aerie-cli activate"
        )
        sys.exit(-1)
    except Exception:
        Console().print_exception()


if __name__ == "__main__":
    main()  # pragma: no cover
